# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Check that a GribMessage can be pickled."""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import pickle
import unittest

from iris.tests.integration.test_pickle import Common as PickleCommon

from iris_grib.message import GribMessage


@tests.skip_data
class TestPickleGribMessage(PickleCommon, tests.IrisGribTest):
    # NOTE: based on 'Common' from iris.tests.integration.test_pickle.
    def setUp(self):
        self.path = tests.get_data_path(("GRIB", "fp_units", "hours.grib2"))

    # NOTE: this looks like it is overriding the parent "pickle_cube", but this
    # function is *not* the same thing.
    # (a) the tests which use pickle_cube are "test_protocol_XXX", and none
    #     of those currently work (see overrides below).
    # (b) the tests which use *this* function are the extra ones below.
    #     These check only that a GribMessage, and its lazy ".data", can be
    #     pickled.  However, still, neither reads back successfully.
    # TODO: resolve this or remove the test.
    def pickle_obj(self, obj):
        with self.temp_filename(".pkl") as filename:
            with open(filename, "wb") as f:
                pickle.dump(obj, f)
            # NOTE: *ought* to read back + check equal, but it does not work.
            # TODO: fix

    # These probably "ought" to work, but currently fail.
    # see https://github.com/SciTools/iris-grib/issues/202
    @unittest.expectedFailure
    def test_protocol_0(self):
        super().test_protocol_0()

    @unittest.expectedFailure
    def test_protocol_1(self):
        super().test_protocol_1()

    @unittest.expectedFailure
    def test_protocol_2(self):
        super().test_protocol_2()

    def test(self):
        # Check that a GribMessage pickles without errors.
        messages = GribMessage.messages_from_filename(self.path)
        obj = next(messages)
        self.pickle_obj(obj)

    def test_data(self):
        # Check that GribMessage.data pickles without errors.
        messages = GribMessage.messages_from_filename(self.path)
        obj = next(messages).data
        self.pickle_obj(obj)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Empty file to allow import.

"""



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for `iris_grib._load_convert.fixup_float32_from_int32`.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from iris_grib._load_convert import fixup_float32_from_int32


class Test(tests.IrisGribTest):
    def test_negative(self):
        result = fixup_float32_from_int32(-0x3f000000)
        self.assertEqual(result, -0.5)

    def test_zero(self):
        result = fixup_float32_from_int32(0)
        self.assertEqual(result, 0)

    def test_positive(self):
        result = fixup_float32_from_int32(0x3f000000)
        self.assertEqual(result, 0.5)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the `iris_grib.GribWrapper` class.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import numpy as np
from unittest import mock

from iris._lazy_data import as_concrete_data, is_lazy_data
from iris.exceptions import TranslationError

from iris_grib import GribWrapper, GribDataProxy, _load_generate


_message_length = 1000


def _mock_grib_get_long(grib_message, key):
    lookup = dict(totalLength=_message_length,
                  numberOfValues=200,
                  jPointsAreConsecutive=0,
                  Ni=20,
                  Nj=10,
                  edition=1)
    try:
        result = lookup[key]
    except KeyError:
        msg = 'Mock grib_get_long unknown key: {!r}'.format(key)
        raise AttributeError(msg)
    return result


def _mock_grib_get_string(grib_message, key):
    return grib_message


def _mock_grib_get_native_type(grib_message, key):
    result = int
    if key == 'gridType':
        result = str
    return result


class Test_edition(tests.IrisGribTest):
    def setUp(self):
        self.patch('iris_grib.GribWrapper._confirm_in_scope')
        self.patch('iris_grib.GribWrapper._compute_extra_keys')
        self.patch('gribapi.grib_get_long', _mock_grib_get_long)
        self.patch('gribapi.grib_get_string', _mock_grib_get_string)
        self.patch('gribapi.grib_get_native_type', _mock_grib_get_native_type)
        self.tell = mock.Mock(side_effect=[_message_length])

    def test_not_edition_1(self):
        def func(grib_message, key):
            return 2

        emsg = "GRIB edition 2 is not supported by 'GribWrapper'"
        with mock.patch('gribapi.grib_get_long', func):
            with self.assertRaisesRegex(TranslationError, emsg):
                GribWrapper(None)

    def test_edition_1(self):
        grib_message = 'regular_ll'
        grib_fh = mock.Mock(tell=self.tell)
        wrapper = GribWrapper(grib_message, grib_fh)
        self.assertEqual(wrapper.grib_message, grib_message)


@tests.skip_data
class Test_deferred_data(tests.IrisTest):
    def test_regular_data(self):
        filename = tests.get_data_path(('GRIB', 'gaussian',
                                        'regular_gg.grib1'))
        messages = list(_load_generate(filename))
        self.assertTrue(is_lazy_data(messages[0]._data))

    def test_reduced_data(self):
        filename = tests.get_data_path(('GRIB', 'reduced',
                                        'reduced_ll.grib1'))
        messages = list(_load_generate(filename))
        self.assertTrue(is_lazy_data(messages[0]._data))


class Test_deferred_proxy_args(tests.IrisTest):
    def setUp(self):
        self.patch('iris_grib.GribWrapper._confirm_in_scope')
        self.patch('iris_grib.GribWrapper._compute_extra_keys')
        self.patch('gribapi.grib_get_long', _mock_grib_get_long)
        self.patch('gribapi.grib_get_string', _mock_grib_get_string)
        self.patch('gribapi.grib_get_native_type', _mock_grib_get_native_type)
        tell_tale = np.arange(1, 5) * _message_length
        self.expected = tell_tale - _message_length
        self.grib_fh = mock.Mock(tell=mock.Mock(side_effect=tell_tale))
        self.dtype = np.float64
        self.path = self.grib_fh.name
        self.lookup = _mock_grib_get_long

    def test_regular_proxy_args(self):
        grib_message = 'regular_ll'
        shape = (self.lookup(grib_message, 'Nj'),
                 self.lookup(grib_message, 'Ni'))
        for offset in self.expected:
            with mock.patch('iris_grib.GribDataProxy') as mock_gdp:
                gw = GribWrapper(grib_message, self.grib_fh)
            mock_gdp.assert_called_once_with(shape, self.dtype,
                                             self.path, offset)

    def test_reduced_proxy_args(self):
        grib_message = 'reduced_gg'
        shape = (self.lookup(grib_message, 'numberOfValues'))
        for offset in self.expected:
            with mock.patch('iris_grib.GribDataProxy') as mock_gdp:
                gw = GribWrapper(grib_message, self.grib_fh)
            mock_gdp.assert_called_once_with((shape,), self.dtype,
                                             self.path, offset)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Integration test for round-trip loading and saving of hybrid height and
hybrid pressure cubes.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from iris import load_cube, load_cubes, save
# Try except allows compatibility with current Iris (2.4) and also master.
# TODO: simplify to just the iris.util import once we drop support for any
# Iris versions with iris.experimental.equalise_cubes import
try:
    from iris.util import equalise_attributes
except ImportError:
    from iris.experimental.equalise_cubes import equalise_attributes


@tests.skip_grib_data
class TestHybridHeightRoundTrip(tests.IrisGribTest):
    def test_hh_round_trip(self):
        filepath = self.get_testdata_path(
            'faked_sample_hh_grib_data.grib2')
        # Load and save temperature cube and reference (orography) cube
        # separately because this is the only way to save the hybrid height
        # coordinate.
        cube, ref_cube = load_cubes(filepath,
                                    ('air_temperature', 'surface_altitude'))

        with self.temp_filename() as tmp_save_path:
            save([cube, ref_cube], tmp_save_path, saver='grib2')
            # Only need to reload temperature cube to compare with unsaved
            # temperature cube.
            saved_cube = load_cube(tmp_save_path, 'air_temperature')
            self.assertTrue(saved_cube == cube)


@tests.skip_grib_data
class TestHybridPressureRoundTrip(tests.IrisGribTest):
    def test_hybrid_pressure(self):
        filepath = self.get_testdata_path(
            'faked_sample_hp_grib_data.grib2')
        # Load and save temperature cube and reference (air_pressure at
        # surface) cube separately because this is the only way to save the
        # hybrid pressure coordinate.
        cube, ref_cube = load_cubes(filepath,
                                    ('air_temperature', 'air_pressure'))

        with self.temp_filename() as tmp_save_path:
            save([cube, ref_cube], tmp_save_path, saver='grib2')
            # Only need to reload temperature cube to compare with unsaved
            # temperature cube.
            saved_cube = load_cube(tmp_save_path, 'air_temperature')

            # Currently all attributes are lost when saving to grib, so we must
            # equalise them in order to successfully compare all other aspects.
            equalise_attributes([saved_cube, cube])

            self.assertTrue(saved_cube == cube)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :meth:`iris_grib._save_rules.grid_definition_template_1`.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import numpy as np

from iris.coord_systems import GeogCS, RotatedGeogCS
from iris.exceptions import TranslationError
from iris.fileformats.pp import EARTH_RADIUS as PP_DEFAULT_EARTH_RADIUS

from iris_grib._save_rules import grid_definition_template_1
from iris_grib.tests.unit.save_rules import GdtTestMixin


class Test(tests.IrisGribTest, GdtTestMixin):
    def setUp(self):
        self.default_ellipsoid = GeogCS(PP_DEFAULT_EARTH_RADIUS)
        GdtTestMixin.setUp(self)

    def _default_coord_system(self):
        # Define an alternate, rotated coordinate system to test.
        cs = RotatedGeogCS(grid_north_pole_latitude=90.0,
                           grid_north_pole_longitude=0.0,
                           ellipsoid=self.default_ellipsoid)
        return cs

    def test__template_number(self):
        grid_definition_template_1(self.test_cube, self.mock_grib)
        self._check_key('gridDefinitionTemplateNumber', 1)

    def test__shape_of_earth_spherical(self):
        ellipsoid = GeogCS(1.23)
        cs = RotatedGeogCS(grid_north_pole_latitude=90.0,
                           grid_north_pole_longitude=0.0,
                           ellipsoid=ellipsoid)
        test_cube = self._make_test_cube(cs=cs)
        grid_definition_template_1(test_cube, self.mock_grib)
        self._check_key('shapeOfTheEarth', 1)
        self._check_key('scaleFactorOfRadiusOfSphericalEarth', 0)
        self._check_key('scaledValueOfRadiusOfSphericalEarth', 1.23)

    def test__shape_of_earth_flattened(self):
        ellipsoid = GeogCS(semi_major_axis=1.456, semi_minor_axis=1.123)
        cs = RotatedGeogCS(grid_north_pole_latitude=90.0,
                           grid_north_pole_longitude=0.0,
                           ellipsoid=ellipsoid)
        test_cube = self._make_test_cube(cs=cs)
        grid_definition_template_1(test_cube, self.mock_grib)
        self._check_key('shapeOfTheEarth', 7)
        self._check_key('scaleFactorOfEarthMajorAxis', 0)
        self._check_key('scaledValueOfEarthMajorAxis', 1.456)
        self._check_key('scaleFactorOfEarthMinorAxis', 0)
        self._check_key('scaledValueOfEarthMinorAxis', 1.123)

    def test__grid_shape(self):
        test_cube = self._make_test_cube(x_points=np.arange(13),
                                         y_points=np.arange(6))
        grid_definition_template_1(test_cube, self.mock_grib)
        self._check_key('Ni', 13)
        self._check_key('Nj', 6)

    def test__grid_points(self):
        test_cube = self._make_test_cube(x_points=[1, 3, 5, 7],
                                         y_points=[4, 9])
        grid_definition_template_1(test_cube, self.mock_grib)
        self._check_key("longitudeOfFirstGridPoint", 1000000)
        self._check_key("longitudeOfLastGridPoint", 7000000)
        self._check_key("latitudeOfFirstGridPoint", 4000000)
        self._check_key("latitudeOfLastGridPoint", 9000000)
        self._check_key("iDirectionIncrement", 2000000)
        self._check_key("jDirectionIncrement", 5000000)

    def test__scanmode(self):
        grid_definition_template_1(self.test_cube, self.mock_grib)
        self._check_key('iScansPositively', 1)
        self._check_key('jScansPositively', 1)

    def test__scanmode_reverse(self):
        test_cube = self._make_test_cube(x_points=np.arange(7, 0, -1))
        grid_definition_template_1(test_cube, self.mock_grib)
        self._check_key('iScansPositively', 0)
        self._check_key('jScansPositively', 1)

    def test__rotated_pole(self):
        cs = RotatedGeogCS(grid_north_pole_latitude=75.3,
                           grid_north_pole_longitude=54.321,
                           ellipsoid=self.default_ellipsoid)
        test_cube = self._make_test_cube(cs=cs)
        grid_definition_template_1(test_cube, self.mock_grib)
        self._check_key("latitudeOfSouthernPole", -75300000)
        self._check_key("longitudeOfSouthernPole", 234321000)
        self._check_key("angleOfRotation", 0)

    def test__fail_rotated_pole_nonstandard_meridian(self):
        cs = RotatedGeogCS(grid_north_pole_latitude=90.0,
                           grid_north_pole_longitude=0.0,
                           north_pole_grid_longitude=22.5,
                           ellipsoid=self.default_ellipsoid)
        test_cube = self._make_test_cube(cs=cs)
        with self.assertRaisesRegex(
                TranslationError,
                'not yet support .* rotated prime meridian.'):
            grid_definition_template_1(test_cube, self.mock_grib)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:func:`iris_grib._load_convert.grid_definition_template_140`.
"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import cartopy.crs as ccrs
import numpy as np

import iris.coord_systems
import iris.coords

from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib._load_convert import _MDI as MDI

from iris_grib._load_convert import grid_definition_template_140


class Test(tests.IrisGribTest):

    def section_3(self):
        section = {
            'gridDefinitionTemplateNumber': 140,
            'shapeOfTheEarth': 4,
            'scaleFactorOfRadiusOfSphericalEarth': MDI,
            'scaledValueOfRadiusOfSphericalEarth': MDI,
            'scaleFactorOfEarthMajorAxis': MDI,
            'scaledValueOfEarthMajorAxis': MDI,
            'scaleFactorOfEarthMinorAxis': MDI,
            'scaledValueOfEarthMinorAxis': MDI,
            'numberOfPointsAlongXAxis': 2,
            'numberOfPointsAlongYAxis': 2,
            'latitudeOfFirstGridPoint': 53988880,
            'longitudeOfFirstGridPoint': -4027984,
            'standardParallelInMicrodegrees': 54900000,
            'centralLongitudeInMicrodegrees': -2500000,
            'resolutionAndComponentFlags': 0b00110000,
            'xDirectionGridLengthInMillimetres': 2000000,
            'yDirectionGridLengthInMillimetres': 2000000,
            'scanningMode': 0b01000000,
        }
        return section

    def expected(self, y_dim, x_dim):
        # Prepare the expectation.
        expected = empty_metadata()
        ellipsoid = iris.coord_systems.GeogCS(
            6378137,
            inverse_flattening=298.257222101)
        cs = iris.coord_systems.LambertAzimuthalEqualArea(
            latitude_of_projection_origin=54.9,
            longitude_of_projection_origin=-2.5,
            false_easting=0,
            false_northing=0,
            ellipsoid=ellipsoid)
        lon0 = -4027984 * 1e-6
        lat0 = 53988880 * 1e-6
        x0m, y0m = cs.as_cartopy_crs().transform_point(
            lon0, lat0, ccrs.Geodetic())
        dxm = dym = 2000.
        x_points = x0m + dxm * np.arange(2)
        y_points = y0m + dym * np.arange(2)
        x = iris.coords.DimCoord(x_points,
                                 standard_name='projection_x_coordinate',
                                 units='m',
                                 coord_system=cs)
        y = iris.coords.DimCoord(y_points,
                                 standard_name='projection_y_coordinate',
                                 units='m',
                                 coord_system=cs)
        expected['dim_coords_and_dims'].append((y, y_dim))
        expected['dim_coords_and_dims'].append((x, x_dim))
        return expected

    def test(self):
        section = self.section_3()
        metadata = empty_metadata()
        grid_definition_template_140(section, metadata)
        expected = self.expected(0, 1)
        self.assertEqual(metadata, expected)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for `iris_grib.message.Section`.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import gribapi
import numpy as np

from iris_grib.message import Section


@tests.skip_data
class Test___getitem__(tests.IrisGribTest):
    def setUp(self):
        filename = tests.get_data_path(('GRIB', 'uk_t', 'uk_t.grib2'))
        with open(filename, 'rb') as grib_fh:
            self.grib_id = gribapi.grib_new_from_file(grib_fh)

    def test_scalar(self):
        section = Section(self.grib_id, None, ['Ni'])
        self.assertEqual(section['Ni'], 47)

    def test_array(self):
        section = Section(self.grib_id, None, ['codedValues'])
        codedValues = section['codedValues']
        self.assertEqual(codedValues.shape, (1551,))
        self.assertArrayAlmostEqual(codedValues[:3],
                                    [-1.78140259, -1.53140259, -1.28140259])

    def test_typeOfFirstFixedSurface(self):
        section = Section(self.grib_id, None, ['typeOfFirstFixedSurface'])
        self.assertEqual(section['typeOfFirstFixedSurface'], 100)

    def test_numberOfSection(self):
        n = 4
        section = Section(self.grib_id, n, ['numberOfSection'])
        self.assertEqual(section['numberOfSection'], n)

    def test_invalid(self):
        section = Section(self.grib_id, None, ['Ni'])
        with self.assertRaisesRegex(KeyError, 'Nii'):
            section['Nii']


@tests.skip_data
class Test__getitem___pdt_31(tests.IrisGribTest):
    def setUp(self):
        filename = tests.get_data_path(('GRIB', 'umukv', 'ukv_chan9.grib2'))
        with open(filename, 'rb') as grib_fh:
            self.grib_id = gribapi.grib_new_from_file(grib_fh)
        self.keys = ['satelliteSeries', 'satelliteNumber', 'instrumentType',
                     'scaleFactorOfCentralWaveNumber',
                     'scaledValueOfCentralWaveNumber']

    def test_array(self):
        section = Section(self.grib_id, None, self.keys)
        for key in self.keys:
            value = section[key]
            self.assertIsInstance(value, np.ndarray)
            self.assertEqual(value.shape, (1,))


@tests.skip_data
class Test_get_computed_key(tests.IrisGribTest):
    def test_gdt40_computed(self):
        fname = tests.get_data_path(('GRIB', 'gaussian', 'regular_gg.grib2'))
        with open(fname, 'rb') as grib_fh:
            self.grib_id = gribapi.grib_new_from_file(grib_fh)
            section = Section(self.grib_id, None, [])
        latitudes = section.get_computed_key('latitudes')
        self.assertTrue(88.55 < latitudes[0] < 88.59)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :meth:`iris_grib._save_rules.grid_definition_template_30`.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import numpy as np

import iris.coords
from iris.coord_systems import GeogCS, LambertConformal
from iris.exceptions import TranslationError

from iris_grib._save_rules import grid_definition_template_30
from iris_grib.tests.unit.save_rules import GdtTestMixin


class FakeGribError(Exception):
    pass


class Test(tests.IrisGribTest, GdtTestMixin):
    def setUp(self):
        self.default_ellipsoid = GeogCS(semi_major_axis=6377563.396,
                                        semi_minor_axis=6356256.909)
        self.test_cube = self._make_test_cube()

        GdtTestMixin.setUp(self)

    def _make_test_cube(self, cs=None, x_points=None, y_points=None):
        # Create a cube with given properties, or minimal defaults.
        if cs is None:
            cs = self._default_coord_system()
        if x_points is None:
            x_points = self._default_x_points()
        if y_points is None:
            y_points = self._default_y_points()

        x_coord = iris.coords.DimCoord(x_points, 'projection_x_coordinate',
                                       units='m', coord_system=cs)
        y_coord = iris.coords.DimCoord(y_points, 'projection_y_coordinate',
                                       units='m', coord_system=cs)
        test_cube = iris.cube.Cube(np.zeros((len(y_points), len(x_points))))
        test_cube.add_dim_coord(y_coord, 0)
        test_cube.add_dim_coord(x_coord, 1)
        return test_cube

    def _default_coord_system(self):
        return LambertConformal(central_lat=39.0, central_lon=-96.0,
                                false_easting=0.0, false_northing=0.0,
                                secant_latitudes=(33, 45),
                                ellipsoid=self.default_ellipsoid)

    def test__template_number(self):
        grid_definition_template_30(self.test_cube, self.mock_grib)
        self._check_key('gridDefinitionTemplateNumber', 30)

    def test__shape_of_earth(self):
        grid_definition_template_30(self.test_cube, self.mock_grib)
        self._check_key('shapeOfTheEarth', 7)
        self._check_key('scaleFactorOfEarthMajorAxis', 0)
        self._check_key('scaledValueOfEarthMajorAxis', 6377563.396)
        self._check_key('scaleFactorOfEarthMinorAxis', 0)
        self._check_key('scaledValueOfEarthMinorAxis', 6356256.909)

    def test__grid_shape(self):
        test_cube = self._make_test_cube(x_points=np.arange(13),
                                         y_points=np.arange(6))
        grid_definition_template_30(test_cube, self.mock_grib)
        self._check_key('Nx', 13)
        self._check_key('Ny', 6)

    def test__grid_points(self):
        test_cube = self._make_test_cube(x_points=[1e6, 3e6, 5e6, 7e6],
                                         y_points=[4e6, 9e6])
        grid_definition_template_30(test_cube, self.mock_grib)
        self._check_key("latitudeOfFirstGridPoint", 71676530)
        self._check_key("longitudeOfFirstGridPoint", 287218188)
        self._check_key("Dx", 2e9)
        self._check_key("Dy", 5e9)

    def test__template_specifics(self):
        grid_definition_template_30(self.test_cube, self.mock_grib)
        self._check_key("LaD", 39e6)
        self._check_key("LoV", 264e6)
        self._check_key("Latin1", 33e6)
        self._check_key("Latin2", 45e6)
        self._check_key("latitudeOfSouthernPole", 0)
        self._check_key("longitudeOfSouthernPole", 0)

    def test__scanmode(self):
        grid_definition_template_30(self.test_cube, self.mock_grib)
        self._check_key('iScansPositively', 1)
        self._check_key('jScansPositively', 1)

    def test__scanmode_reverse(self):
        test_cube = self._make_test_cube(x_points=np.arange(7e6, 0, -1e6))
        grid_definition_template_30(test_cube, self.mock_grib)
        self._check_key('iScansPositively', 0)
        self._check_key('jScansPositively', 1)

    def test_projection_centre(self):
        grid_definition_template_30(self.test_cube, self.mock_grib)
        self._check_key("projectionCentreFlag", 0)

    def test_projection_centre_south_pole(self):
        cs = LambertConformal(central_lat=39.0, central_lon=-96.0,
                              false_easting=0.0, false_northing=0.0,
                              secant_latitudes=(-33, -45),
                              ellipsoid=self.default_ellipsoid)
        test_cube = self._make_test_cube(cs=cs)
        grid_definition_template_30(test_cube, self.mock_grib)
        self._check_key("projectionCentreFlag", 1)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for `iris_grib.grib_save_rules.reference_time`."""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock

import gribapi

from iris_grib import load_cubes
from iris_grib._save_rules import reference_time


class Test(tests.IrisGribTest):
    def _test(self, cube):
        grib = mock.Mock()
        mock_gribapi = mock.Mock(spec=gribapi)
        with mock.patch('iris_grib._save_rules.gribapi', mock_gribapi):
            reference_time(cube, grib)

        mock_gribapi.assert_has_calls(
            [mock.call.grib_set_long(grib, "significanceOfReferenceTime", 1),
             mock.call.grib_set_long(grib, "dataDate", '19980306'),
             mock.call.grib_set_long(grib, "dataTime", '0300')])

    @tests.skip_data
    def test_forecast_period(self):
        # The stock cube has a non-compliant forecast_period.
        fname = tests.get_data_path(('GRIB', 'global_t', 'global.grib2'))
        [cube] = load_cubes(fname)
        self._test(cube)

    @tests.skip_data
    def test_no_forecast_period(self):
        # The stock cube has a non-compliant forecast_period.
        fname = tests.get_data_path(('GRIB', 'global_t', 'global.grib2'))
        [cube] = load_cubes(fname)
        cube.remove_coord("forecast_period")
        self._test(cube)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:func:`iris_grib._load_convert.grid_definition_template_10`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import numpy as np

import iris.coord_systems
import iris.coords
import iris.exceptions

from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib._load_convert import _MDI as MDI

from iris_grib._load_convert import grid_definition_template_10


class Test(tests.IrisGribTest):
    def section_3(self):
        section = {
            'gridDefinitionTemplateNumber': 10,
            'shapeOfTheEarth': 1,
            'scaleFactorOfRadiusOfSphericalEarth': 0,
            'scaledValueOfRadiusOfSphericalEarth': 6371200,
            'scaleFactorOfEarthMajorAxis': 0,
            'scaledValueOfEarthMajorAxis': 0,
            'scaleFactorOfEarthMinorAxis': 0,
            'scaledValueOfEarthMinorAxis': 0,
            'Ni': 181,
            'Nj': 213,
            'latitudeOfFirstGridPoint': 2351555,
            'latitudeOfLastGridPoint': 25088204,
            'LaD': 14000000,
            'longitudeOfFirstGridPoint': 114990304,
            'longitudeOfLastGridPoint': 135009712,
            'resolutionAndComponentFlags': 56,
            'scanningMode': 64,
            'Di': 12000000,
            'Dj': 12000000
        }
        return section

    def expected(self, y_dim, x_dim):
        # Prepare the expectation.
        expected = empty_metadata()
        ellipsoid = iris.coord_systems.GeogCS(6371200.0)
        cs = iris.coord_systems.Mercator(standard_parallel=14.,
                                         ellipsoid=ellipsoid)
        nx = 181
        x_origin = 12406918.990644248
        dx = 12000
        x = iris.coords.DimCoord(np.arange(nx) * dx + x_origin,
                                 'projection_x_coordinate',
                                 units='m',
                                 coord_system=cs)
        ny = 213
        y_origin = 253793.10903714446
        dy = 12000
        y = iris.coords.DimCoord(np.arange(ny) * dy + y_origin,
                                 'projection_y_coordinate',
                                 units='m',
                                 coord_system=cs)
        expected['dim_coords_and_dims'].append((y, y_dim))
        expected['dim_coords_and_dims'].append((x, x_dim))
        return expected

    def test(self):
        section = self.section_3()
        metadata = empty_metadata()
        grid_definition_template_10(section, metadata)
        expected = self.expected(y_dim=0, x_dim=1)
        self.assertEqual(metadata, expected)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Integration tests for round-trip loading and saving various product
definitions.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from cf_units import Unit
from iris import load_cube, save
import iris.coords
import iris.coord_systems

import iris.tests.stock as stock

from iris_grib.grib_phenom_translation import GRIBCode


class TestPDT11(tests.TestGribMessage):
    def test_perturbation(self):
        path = tests.get_data_path(
            ("NetCDF", "global", "xyt", "SMALL_hires_wind_u_for_ipcc4.nc")
        )
        cube = load_cube(path)
        # trim to 1 time and regular lats
        cube = cube[0, 12:144, :]
        crs = iris.coord_systems.GeogCS(6371229)
        cube.coord("latitude").coord_system = crs
        cube.coord("longitude").coord_system = crs
        # add a realization coordinate
        cube.add_aux_coord(
            iris.coords.DimCoord(
                points=1, standard_name="realization", units="1"
            )
        )
        with self.temp_filename("testPDT11.GRIB2") as temp_file_path:
            iris.save(cube, temp_file_path)

            # Check that various aspects of the saved file are as expected.
            expect_values = (
                (0, "editionNumber", 2),
                (3, "gridDefinitionTemplateNumber", 0),
                (4, "productDefinitionTemplateNumber", 11),
                (4, "perturbationNumber", 1),
                (4, "typeOfStatisticalProcessing", 0),
                (4, "numberOfForecastsInEnsemble", 255),
            )
            self.assertGribMessageContents(temp_file_path, expect_values)


class TestPDT40(tests.IrisTest):
    def test_save_load(self):
        cube = stock.lat_lon_cube()
        cube.rename("atmosphere_mole_content_of_ozone")
        cube.units = Unit("Dobson")
        tcoord = iris.coords.DimCoord(
            23, "time", units=Unit("days since epoch", calendar="standard")
        )
        fpcoord = iris.coords.DimCoord(24, "forecast_period",
                                       units=Unit("hours"))
        cube.add_aux_coord(tcoord)
        cube.add_aux_coord(fpcoord)
        cube.attributes["WMO_constituent_type"] = 0
        cube.attributes["GRIB_PARAM"] = GRIBCode("GRIB2:d000c014n000")

        with self.temp_filename("test_grib_pdt40.grib2") as temp_file_path:
            save(cube, temp_file_path)
            loaded = load_cube(temp_file_path)
            self.assertEqual(loaded.attributes, cube.attributes)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris_grib._load_convert.resolution_flags.`

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from iris_grib._load_convert import resolution_flags, ResolutionFlags


class Test(tests.IrisGribTest):
    def test_unset(self):
        expected = ResolutionFlags(False, False, False)
        self.assertEqual(resolution_flags(0x0), expected)

    def test_i_increments_given(self):
        expected = ResolutionFlags(True, False, False)
        self.assertEqual(resolution_flags(0x20), expected)

    def test_j_increments_given(self):
        expected = ResolutionFlags(False, True, False)
        self.assertEqual(resolution_flags(0x10), expected)

    def test_uv_resolved(self):
        expected = ResolutionFlags(False, False, True)
        self.assertEqual(resolution_flags(0x08), expected)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris_grib._save_rules.product_definition_template_40`

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock

from cf_units import Unit
import gribapi

from iris.coords import DimCoord
import iris.tests.stock as stock

from iris_grib._save_rules import product_definition_template_40


class TestChemicalConstituentIdentifier(tests.IrisGribTest):
    def setUp(self):
        self.cube = stock.lat_lon_cube()
        # Rename cube to avoid warning about unknown discipline/parameter.
        self.cube.rename('atmosphere_mole_content_of_ozone')
        coord = DimCoord(24, 'time',
                         units=Unit('days since epoch', calendar='standard'))
        self.cube.add_aux_coord(coord)
        self.cube.attributes['WMO_constituent_type'] = 0

    @mock.patch.object(gribapi, 'grib_set')
    def test_constituent_type(self, mock_set):
        cube = self.cube

        product_definition_template_40(cube, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 'productDefinitionTemplateNumber', 40)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 'constituentType', 0)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:func:`iris_grib._load_convert.grid_definition_template_40`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import numpy as np

import iris.coord_systems
import iris.coords

from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib._load_convert import _MDI as MDI

from iris_grib._load_convert import grid_definition_template_40


class _Section(dict):
    def get_computed_key(self, key):
        return self.get(key)


class Test_regular(tests.IrisGribTest):

    def section_3(self):
        section = _Section({
            'shapeOfTheEarth': 0,
            'scaleFactorOfRadiusOfSphericalEarth': 0,
            'scaledValueOfRadiusOfSphericalEarth': 6367470,
            'scaleFactorOfEarthMajorAxis': 0,
            'scaledValueOfEarthMajorAxis': MDI,
            'scaleFactorOfEarthMinorAxis': 0,
            'scaledValueOfEarthMinorAxis': MDI,
            'iDirectionIncrement': 22500000,
            'longitudeOfFirstGridPoint': 0,
            'resolutionAndComponentFlags': 32,
            'Ni': 16,
            'scanningMode': 0b01000000,
            'distinctLatitudes': np.array([-73.79921363, -52.81294319,
                                           -31.70409175, -10.56988231,
                                           10.56988231,  31.70409175,
                                           52.81294319,  73.79921363]),
            'numberOfOctectsForNumberOfPoints': 0,
            'interpretationOfNumberOfPoints': 0,
        })
        return section

    def expected(self, y_dim, x_dim, y_neg=True):
        # Prepare the expectation.
        expected = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        nx = 16
        dx = 22.5
        x_origin = 0
        x = iris.coords.DimCoord(np.arange(nx) * dx + x_origin,
                                 standard_name='longitude',
                                 units='degrees_east',
                                 coord_system=cs,
                                 circular=True)
        y_points = np.array([73.79921363, 52.81294319,
                             31.70409175, 10.56988231,
                             -10.56988231,  -31.70409175,
                             -52.81294319,  -73.79921363])
        if not y_neg:
            y_points = y_points[::-1]
        y = iris.coords.DimCoord(y_points,
                                 standard_name='latitude',
                                 units='degrees_north',
                                 coord_system=cs)
        expected['dim_coords_and_dims'].append((y, y_dim))
        expected['dim_coords_and_dims'].append((x, x_dim))
        return expected

    def test(self):
        section = self.section_3()
        metadata = empty_metadata()
        grid_definition_template_40(section, metadata)
        expected = self.expected(0, 1, y_neg=False)
        self.assertEqual(metadata, expected)

    def test_transposed(self):
        section = self.section_3()
        section['scanningMode'] = 0b01100000
        metadata = empty_metadata()
        grid_definition_template_40(section, metadata)
        expected = self.expected(1, 0, y_neg=False)
        self.assertEqual(metadata, expected)

    def test_reverse_latitude(self):
        section = self.section_3()
        section['scanningMode'] = 0b00000000
        metadata = empty_metadata()
        grid_definition_template_40(section, metadata)
        expected = self.expected(0, 1, y_neg=True)
        self.assertEqual(metadata, expected)


class Test_reduced(tests.IrisGribTest):

    def section_3(self):
        section = _Section({
            'shapeOfTheEarth': 0,
            'scaleFactorOfRadiusOfSphericalEarth': 0,
            'scaledValueOfRadiusOfSphericalEarth': 6367470,
            'scaleFactorOfEarthMajorAxis': 0,
            'scaledValueOfEarthMajorAxis': MDI,
            'scaleFactorOfEarthMinorAxis': 0,
            'scaledValueOfEarthMinorAxis': MDI,
            'longitudes': np.array([0., 180.,
                                    0., 120., 240.,
                                    0., 120., 240.,
                                    0., 180.]),
            'latitudes': np.array([-59.44440829, -59.44440829,
                                   -19.87571915, -19.87571915, -19.87571915,
                                   19.87571915, 19.87571915, 19.87571915,
                                   59.44440829, 59.44440829]),
            'numberOfOctectsForNumberOfPoints': 1,
            'interpretationOfNumberOfPoints': 1,
        })
        return section

    def expected(self):
        # Prepare the expectation.
        expected = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        x_points = np.array([0., 180.,
                             0., 120., 240.,
                             0., 120., 240.,
                             0., 180.])
        y_points = np.array([-59.44440829, -59.44440829,
                             -19.87571915, -19.87571915, -19.87571915,
                             19.87571915, 19.87571915, 19.87571915,
                             59.44440829, 59.44440829])
        x = iris.coords.AuxCoord(x_points,
                                 standard_name='longitude',
                                 units='degrees_east',
                                 coord_system=cs)
        y = iris.coords.AuxCoord(y_points,
                                 standard_name='latitude',
                                 units='degrees_north',
                                 coord_system=cs)
        expected['aux_coords_and_dims'].append((y, 0))
        expected['aux_coords_and_dims'].append((x, 0))
        return expected

    def test(self):
        section = self.section_3()
        metadata = empty_metadata()
        expected = self.expected()
        grid_definition_template_40(section, metadata)
        self.assertEqual(metadata, expected)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests for specific implementation aspects of the grib loaders.
Old, and GRIB-1 specific.
Ported here from 'iris.tests.test_grib_load_translations'.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else
import iris_grib.tests as tests

import datetime
from unittest import mock

import cf_units
import numpy as np

import iris
import iris.exceptions

import gribapi
import iris.fileformats
import iris_grib


def _mock_gribapi_fetch(message, key):
    """
    Fake the gribapi key-fetch.

    Fetch key-value from the fake message (dictionary).
    If the key is not present, raise the diagnostic exception.

    """
    if key in message:
        return message[key]
    else:
        raise _mock_gribapi.errors.GribInternalError


def _mock_gribapi__grib_is_missing(grib_message, keyname):
    """
    Fake the gribapi key-existence enquiry.

    Return whether the key exists in the fake message (dictionary).

    """
    return keyname not in grib_message


def _mock_gribapi__grib_get_native_type(grib_message, keyname):
    """
    Fake the gribapi type-discovery operation.

    Return type of key-value in the fake message (dictionary).
    If the key is not present, raise the diagnostic exception.

    """
    if keyname in grib_message:
        return type(grib_message[keyname])
    raise _mock_gribapi.errors.GribInternalError(keyname)


# Construct a mock object to mimic the gribapi for GribWrapper testing.
_mock_gribapi = mock.Mock(spec=gribapi)
_mock_gribapi.errors.GribInternalError = Exception

_mock_gribapi.grib_get_long = mock.Mock(side_effect=_mock_gribapi_fetch)
_mock_gribapi.grib_get_string = mock.Mock(side_effect=_mock_gribapi_fetch)
_mock_gribapi.grib_get_double = mock.Mock(side_effect=_mock_gribapi_fetch)
_mock_gribapi.grib_get_double_array = mock.Mock(
    side_effect=_mock_gribapi_fetch
)
_mock_gribapi.grib_is_missing = mock.Mock(
    side_effect=_mock_gribapi__grib_is_missing
)
_mock_gribapi.grib_get_native_type = mock.Mock(
    side_effect=_mock_gribapi__grib_get_native_type
)

# define seconds in an hour, for general test usage
_hour_secs = 3600.0


class FakeGribMessage(dict):
    """
    A 'fake grib message' object, for testing GribWrapper construction.

    Behaves as a dictionary, containing key-values for message keys.

    """

    def __init__(self, **kwargs):
        """
        Create a fake message object.

        General keys can be set/add as required via **kwargs.
        The keys 'edition' and 'time_code' are specially managed.

        """
        # Start with a bare dictionary
        dict.__init__(self)
        # Extract specially-recognised keys.
        edition = kwargs.pop("edition", 1)
        # This testing is only for old-style Grib-1 code.
        assert edition == 1
        time_code = kwargs.pop("time_code", None)
        # Set the minimally required keys.
        self._init_minimal_message(edition=edition)
        # Also set a time-code, if given.
        if time_code is not None:
            self.set_timeunit_code(time_code)
        # Finally, add any remaining passed key-values.
        self.update(**kwargs)

    def _init_minimal_message(self, edition=1):
        # Set values for all the required keys.
        # 'edition' controls the edition-specific keys.
        self.update(
            {
                "Ni": 1,
                "Nj": 1,
                "numberOfValues": 1,
                "alternativeRowScanning": 0,
                "centre": "ecmf",
                "year": 2007,
                "month": 3,
                "day": 23,
                "hour": 12,
                "minute": 0,
                "indicatorOfUnitOfTimeRange": 1,
                "shapeOfTheEarth": 6,
                "gridType": "rotated_ll",
                "angleOfRotation": 0.0,
                "iDirectionIncrementInDegrees": 0.036,
                "jDirectionIncrementInDegrees": 0.036,
                "iScansNegatively": 0,
                "jScansPositively": 1,
                "longitudeOfFirstGridPointInDegrees": -5.70,
                "latitudeOfFirstGridPointInDegrees": -4.452,
                "jPointsAreConsecutive": 0,
                "values": np.array([[1.0]]),
                "indicatorOfParameter": 9999,
                "parameterNumber": 9999,
                "startStep": 24,
                "timeRangeIndicator": 1,
                "P1": 2,
                "P2": 0,
                # time unit - needed AS WELL as 'indicatorOfUnitOfTimeRange'
                "unitOfTime": 1,
                "table2Version": 9999,
            }
        )
        # Add edition-dependent settings.
        self["edition"] = edition

    def set_timeunit_code(self, timecode):
        # Do timecode setting (somewhat edition-dependent).
        self["indicatorOfUnitOfTimeRange"] = timecode
        # for some odd reason, GRIB1 code uses *both* of these
        # NOTE kludge -- the 2 keys are really the same thing
        self["unitOfTime"] = timecode


class TestGribTimecodes(tests.IrisTest):
    def _run_timetests(self, test_set):
        # Check the unit-handling for given units-codes and editions.

        # Operates on lists of cases for various time-units and grib-editions.
        # Format: (edition, code, expected-exception,
        #          equivalent-seconds, description-string)
        with mock.patch("iris_grib.gribapi", _mock_gribapi):
            for test_controls in test_set:
                (
                    grib_edition,
                    timeunit_codenum,
                    expected_error,
                    timeunit_secs,
                    timeunit_str,
                ) = test_controls

                # Construct a suitable fake test message.
                message = FakeGribMessage(
                    edition=grib_edition, time_code=timeunit_codenum
                )

                if expected_error:
                    # Expect GribWrapper construction to fail.
                    with self.assertRaises(type(expected_error)) as ar_context:
                        _ = iris_grib.GribWrapper(message)
                    self.assertEqual(
                        ar_context.exception.args, expected_error.args
                    )
                    continue

                # 'ELSE'...
                # Expect the wrapper construction to work.
                # Make a GribWrapper object and test it.
                wrapped_msg = iris_grib.GribWrapper(message)

                # Check the units string.
                forecast_timeunit = wrapped_msg._forecastTimeUnit
                self.assertEqual(
                    forecast_timeunit,
                    timeunit_str,
                    "Bad unit string for edition={ed:01d}, "
                    "unitcode={code:01d} : "
                    'expected="{wanted}" GOT="{got}"'.format(
                        ed=grib_edition,
                        code=timeunit_codenum,
                        wanted=timeunit_str,
                        got=forecast_timeunit,
                    ),
                )

                # Check the data-starttime calculation.
                interval_start_to_end = (
                    wrapped_msg._phenomenonDateTime -
                    wrapped_msg._referenceDateTime
                )
                if grib_edition == 1:
                    interval_from_units = wrapped_msg.P1
                else:
                    interval_from_units = wrapped_msg.forecastTime
                interval_from_units *= datetime.timedelta(0, timeunit_secs)
                self.assertEqual(
                    interval_start_to_end,
                    interval_from_units,
                    "Inconsistent start time offset for edition={ed:01d}, "
                    "unitcode={code:01d} : "
                    'from-unit="{unit_str}" '
                    'from-phenom-minus-ref="{e2e_str}"'.format(
                        ed=grib_edition,
                        code=timeunit_codenum,
                        unit_str=interval_from_units,
                        e2e_str=interval_start_to_end,
                    ),
                )

    # Test groups of testcases for various time-units and grib-editions.
    # Format: (edition, code, expected-exception,
    #          equivalent-seconds, description-string)
    def test_timeunits_common(self):
        tests = (
            (1, 0, None, 60.0, "minutes"),
            (1, 1, None, _hour_secs, "hours"),
            (1, 2, None, 24.0 * _hour_secs, "days"),
            (1, 10, None, 3.0 * _hour_secs, "3 hours"),
            (1, 11, None, 6.0 * _hour_secs, "6 hours"),
            (1, 12, None, 12.0 * _hour_secs, "12 hours"),
        )
        TestGribTimecodes._run_timetests(self, tests)

    @staticmethod
    def _err_bad_timeunit(code):
        return iris.exceptions.NotYetImplementedError(
            "Unhandled time unit for forecast "
            "indicatorOfUnitOfTimeRange : {code}".format(code=code)
        )

    def test_timeunits_grib1_specific(self):
        tests = (
            (1, 13, None, 0.25 * _hour_secs, "15 minutes"),
            (1, 14, None, 0.5 * _hour_secs, "30 minutes"),
            (1, 254, None, 1.0, "seconds"),
            (1, 111, TestGribTimecodes._err_bad_timeunit(111), 1.0, "??"),
        )
        TestGribTimecodes._run_timetests(self, tests)

    def test_timeunits_calendar(self):
        tests = (
            (1, 3, TestGribTimecodes._err_bad_timeunit(3), 0.0, "months"),
            (1, 4, TestGribTimecodes._err_bad_timeunit(4), 0.0, "years"),
            (1, 5, TestGribTimecodes._err_bad_timeunit(5), 0.0, "decades"),
            (1, 6, TestGribTimecodes._err_bad_timeunit(6), 0.0, "30 years"),
            (1, 7, TestGribTimecodes._err_bad_timeunit(7), 0.0, "centuries"),
        )
        TestGribTimecodes._run_timetests(self, tests)

    def test_timeunits_invalid(self):
        tests = (
            (1, 111, TestGribTimecodes._err_bad_timeunit(111), 1.0, "??"),
        )
        TestGribTimecodes._run_timetests(self, tests)

    def test_warn_unknown_pdts(self):
        # Test loading of an unrecognised GRIB Product Definition Template.

        # Get a temporary file by name (deleted afterward by context).
        with self.temp_filename() as temp_gribfile_path:
            # Write a test grib message to the temporary file.
            with open(temp_gribfile_path, "wb") as temp_gribfile:
                grib_message = gribapi.grib_new_from_samples("GRIB2")
                # Set the PDT to something unexpected.
                gribapi.grib_set_long(
                    grib_message, "productDefinitionTemplateNumber", 5
                )
                gribapi.grib_write(grib_message, temp_gribfile)

            # Load the message from the file as a cube.
            cube_generator = iris_grib.load_cubes(temp_gribfile_path)
            with self.assertRaises(iris.exceptions.TranslationError) as te:
                _ = next(cube_generator)
            self.assertEqual(
                "Product definition template [5]" " is not supported",
                str(te.exception),
            )


class TestGrib1LoadPhenomenon(tests.IrisTest):
    # Test recognition of grib phenomenon types.
    def mock_grib(self):
        grib = mock.Mock()
        grib.edition = 1
        grib.startStep = 0
        grib.phenomenon_points = lambda unit: 3
        grib._forecastTimeUnit = "hours"
        grib.productDefinitionTemplateNumber = 0
        # define a level type (NB these 2 are effectively the same)
        grib.levelType = 1
        grib.typeOfFirstFixedSurface = 1
        grib.typeOfSecondFixedSurface = 1
        return grib

    def cube_from_message(self, grib):
        # Parameter translation now uses the GribWrapper, so we must convert
        # the Mock-based fake message to a FakeGribMessage.
        with mock.patch("iris_grib.gribapi", _mock_gribapi):
            grib_message = FakeGribMessage(**grib.__dict__)
            wrapped_msg = iris_grib.GribWrapper(grib_message)
            cube, _, _ = iris.fileformats.rules._make_cube(
                wrapped_msg, iris_grib._grib1_load_rules.grib1_convert
            )
        return cube

    def test_grib1_unknownparam(self):
        grib = self.mock_grib()
        grib.table2Version = 0
        grib.indicatorOfParameter = 9999
        cube = self.cube_from_message(grib)
        self.assertEqual(cube.standard_name, None)
        self.assertEqual(cube.long_name, None)
        self.assertEqual(cube.units, cf_units.Unit("???"))

    def test_grib1_unknown_local_param(self):
        grib = self.mock_grib()
        grib.table2Version = 128
        grib.indicatorOfParameter = 999
        cube = self.cube_from_message(grib)
        self.assertEqual(cube.standard_name, None)
        self.assertEqual(cube.long_name, "UNKNOWN LOCAL PARAM 999.128")
        self.assertEqual(cube.units, cf_units.Unit("???"))

    def test_grib1_unknown_standard_param(self):
        grib = self.mock_grib()
        grib.table2Version = 1
        grib.indicatorOfParameter = 975
        cube = self.cube_from_message(grib)
        self.assertEqual(cube.standard_name, None)
        self.assertEqual(cube.long_name, "UNKNOWN LOCAL PARAM 975.1")
        self.assertEqual(cube.units, cf_units.Unit("???"))

    def known_grib1(self, param, standard_str, units_str):
        grib = self.mock_grib()
        grib.table2Version = 1
        grib.indicatorOfParameter = param
        cube = self.cube_from_message(grib)
        self.assertEqual(cube.standard_name, standard_str)
        self.assertEqual(cube.long_name, None)
        self.assertEqual(cube.units, cf_units.Unit(units_str))

    def test_grib1_known_standard_params(self):
        # at present, there are just a very few of these
        self.known_grib1(11, "air_temperature", "kelvin")
        self.known_grib1(33, "x_wind", "m s-1")
        self.known_grib1(34, "y_wind", "m s-1")


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests for function
:func:`iris_grib._load_convert.product_definition_template_9`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from unittest import mock

from iris.exceptions import TranslationError

from iris_grib._load_convert import product_definition_template_9
from iris_grib._load_convert import Probability, _MDI


class Test(tests.IrisGribTest):
    def setUp(self):
        # Create patches for called routines
        module = 'iris_grib._load_convert'
        self.patch_pdt8_call = self.patch(
            module + '.product_definition_template_8')
        # Construct dummy call arguments
        self.section = {}
        self.section['probabilityType'] = 1
        self.section['scaledValueOfUpperLimit'] = 53
        self.section['scaleFactorOfUpperLimit'] = 1
        self.frt_coord = mock.sentinel.frt_coord
        self.metadata = {'cell_methods': [mock.sentinel.cell_method],
                         'aux_coords_and_dims': []}

    def test_basic(self):
        result = product_definition_template_9(
            self.section, self.metadata, self.frt_coord)
        # Check expected function was called.
        self.assertEqual(
            self.patch_pdt8_call.call_args_list,
            [mock.call(self.section, self.metadata, self.frt_coord)])
        # Check metadata content (N.B. cell_method has been removed!).
        self.assertEqual(self.metadata, {'cell_methods': [],
                                         'aux_coords_and_dims': []})
        # Check result.
        self.assertEqual(result, Probability('above_threshold', 5.3))

    def test_fail_bad_probability_type(self):
        self.section['probabilityType'] = 17
        with self.assertRaisesRegex(TranslationError,
                                    'unsupported probability type'):
            product_definition_template_9(
                self.section, self.metadata, self.frt_coord)

    def test_fail_bad_threshold_value(self):
        self.section['scaledValueOfUpperLimit'] = _MDI
        with self.assertRaisesRegex(TranslationError,
                                    'missing scaled value'):
            product_definition_template_9(
                self.section, self.metadata, self.frt_coord)

    def test_fail_bad_threshold_scalefactor(self):
        self.section['scaleFactorOfUpperLimit'] = _MDI
        with self.assertRaisesRegex(TranslationError,
                                    'missing scale factor'):
            product_definition_template_9(
                self.section, self.metadata, self.frt_coord)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# Historically this was auto-generated from
# SciTools/iris-code-generators:tools/gen_rules.py

import warnings

from cf_units import CALENDAR_GREGORIAN, Unit
import numpy as np

from iris.aux_factory import HybridPressureFactory
from iris.coords import AuxCoord, CellMethod, DimCoord
from iris.exceptions import TranslationError
from iris.fileformats.rules import (ConversionMetadata, Factory, Reference,
                                    ReferenceTarget)


def grib1_convert(grib):
    """
    Converts a GRIB1 message into the corresponding items of Cube metadata.

    Args:

    * grib:
        A :class:`~iris_grib.GribWrapper` object.

    Returns:
        A :class:`iris.fileformats.rules.ConversionMetadata` object.

    """
    if grib.edition != 1:
        emsg = 'GRIB edition {} is not supported by {!r}.'
        raise TranslationError(emsg.format(grib.edition,
                                           type(grib).__name__))

    factories = []
    references = []
    standard_name = None
    long_name = None
    units = None
    attributes = {}
    cell_methods = []
    dim_coords_and_dims = []
    aux_coords_and_dims = []

    if \
            (grib.gridType=="reduced_gg"):
        aux_coords_and_dims.append((AuxCoord(grib._y_points, grib._y_coord_name, units='degrees', coord_system=grib._coord_system), 0))
        aux_coords_and_dims.append((AuxCoord(grib._x_points, grib._x_coord_name, units='degrees', coord_system=grib._coord_system), 0))

    if \
            (grib.gridType=="regular_ll") and \
            (grib.jPointsAreConsecutive == 0):
        dim_coords_and_dims.append((DimCoord(grib._y_points, grib._y_coord_name, units='degrees', coord_system=grib._coord_system), 0))
        dim_coords_and_dims.append((DimCoord(grib._x_points, grib._x_coord_name, units='degrees', coord_system=grib._coord_system, circular=grib._x_circular), 1))

    if \
            (grib.gridType=="regular_ll") and \
            (grib.jPointsAreConsecutive == 1):
        dim_coords_and_dims.append((DimCoord(grib._y_points, grib._y_coord_name, units='degrees', coord_system=grib._coord_system), 1))
        dim_coords_and_dims.append((DimCoord(grib._x_points, grib._x_coord_name, units='degrees', coord_system=grib._coord_system, circular=grib._x_circular), 0))

    if \
            (grib.gridType=="regular_gg") and \
            (grib.jPointsAreConsecutive == 0):
        dim_coords_and_dims.append((DimCoord(grib._y_points, grib._y_coord_name, units='degrees', coord_system=grib._coord_system), 0))
        dim_coords_and_dims.append((DimCoord(grib._x_points, grib._x_coord_name, units='degrees', coord_system=grib._coord_system, circular=grib._x_circular), 1))

    if \
            (grib.gridType=="regular_gg") and \
            (grib.jPointsAreConsecutive == 1):
        dim_coords_and_dims.append((DimCoord(grib._y_points, grib._y_coord_name, units='degrees', coord_system=grib._coord_system), 1))
        dim_coords_and_dims.append((DimCoord(grib._x_points, grib._x_coord_name, units='degrees', coord_system=grib._coord_system, circular=grib._x_circular), 0))

    if \
            (grib.gridType=="rotated_ll") and \
            (grib.jPointsAreConsecutive == 0):
        dim_coords_and_dims.append((DimCoord(grib._y_points, grib._y_coord_name, units='degrees', coord_system=grib._coord_system), 0))
        dim_coords_and_dims.append((DimCoord(grib._x_points, grib._x_coord_name, units='degrees', coord_system=grib._coord_system, circular=grib._x_circular), 1))

    if \
            (grib.gridType=="rotated_ll") and \
            (grib.jPointsAreConsecutive == 1):
        dim_coords_and_dims.append((DimCoord(grib._y_points, grib._y_coord_name, units='degrees', coord_system=grib._coord_system), 1))
        dim_coords_and_dims.append((DimCoord(grib._x_points, grib._x_coord_name, units='degrees', coord_system=grib._coord_system, circular=grib._x_circular), 0))

    if grib.gridType in ["polar_stereographic", "lambert"]:
        dim_coords_and_dims.append((DimCoord(grib._y_points, grib._y_coord_name, units="m", coord_system=grib._coord_system), 0))
        dim_coords_and_dims.append((DimCoord(grib._x_points, grib._x_coord_name, units="m", coord_system=grib._coord_system), 1))

    if \
            (grib.table2Version < 128) and \
            (grib.indicatorOfParameter == 11) and \
            (grib._cf_data is None):
        standard_name = "air_temperature"
        units = "kelvin"

    if \
            (grib.table2Version < 128) and \
            (grib.indicatorOfParameter == 33) and \
            (grib._cf_data is None):
        standard_name = "x_wind"
        units = "m s-1"

    if \
            (grib.table2Version < 128) and \
            (grib.indicatorOfParameter == 34) and \
            (grib._cf_data is None):
        standard_name = "y_wind"
        units = "m s-1"

    if \
            (grib._cf_data is not None):
        standard_name = grib._cf_data.standard_name
        long_name = grib._cf_data.standard_name or grib._cf_data.long_name
        units = grib._cf_data.units

    if \
            (grib.table2Version >= 128) and \
            (grib._cf_data is None):
        long_name = "UNKNOWN LOCAL PARAM " + str(grib.indicatorOfParameter) + "." + str(grib.table2Version)
        units = "???"

    if \
            (grib.table2Version == 1) and \
            (grib.indicatorOfParameter >= 128):
        long_name = "UNKNOWN LOCAL PARAM " + str(grib.indicatorOfParameter) + "." + str(grib.table2Version)
        units = "???"

    if \
            (grib._phenomenonDateTime != -1.0):
        aux_coords_and_dims.append((DimCoord(points=grib.startStep, standard_name='forecast_period', units=grib._forecastTimeUnit), None))
        aux_coords_and_dims.append((DimCoord(points=grib.phenomenon_points('hours'), standard_name='time', units=Unit('hours since epoch', CALENDAR_GREGORIAN)), None))

    def add_bounded_time_coords(aux_coords_and_dims, grib):
        t_bounds = grib.phenomenon_bounds('hours')
        period = Unit('hours').convert(t_bounds[1] - t_bounds[0],
                                       grib._forecastTimeUnit)
        aux_coords_and_dims.append((
            DimCoord(standard_name='forecast_period',
                     units=grib._forecastTimeUnit,
                     points=grib._forecastTime + 0.5 * period,
                     bounds=[grib._forecastTime, grib._forecastTime + period]),
            None))
        aux_coords_and_dims.append((
            DimCoord(standard_name='time',
                     units=Unit('hours since epoch', CALENDAR_GREGORIAN),
                     points=0.5 * (t_bounds[0] + t_bounds[1]),
                     bounds=t_bounds),
            None))

    if \
            (grib.timeRangeIndicator == 2):
        add_bounded_time_coords(aux_coords_and_dims, grib)

    if \
            (grib.timeRangeIndicator == 3):
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("mean", coords="time"))

    if \
            (grib.timeRangeIndicator == 4):
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("sum", coords="time"))

    if \
            (grib.timeRangeIndicator == 5):
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("_difference", coords="time"))

    if \
            (grib.timeRangeIndicator == 51):
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("mean", coords="time"))

    if \
            (grib.timeRangeIndicator == 113):
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("mean", coords="time"))

    if \
            (grib.timeRangeIndicator == 114):
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("sum", coords="time"))

    if \
            (grib.timeRangeIndicator == 115):
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("mean", coords="time"))

    if \
            (grib.timeRangeIndicator == 116):
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("sum", coords="time"))

    if \
            (grib.timeRangeIndicator == 117):
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("mean", coords="time"))

    if \
            (grib.timeRangeIndicator == 118):
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("_covariance", coords="time"))

    if \
            (grib.timeRangeIndicator == 123):
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("mean", coords="time"))

    if \
            (grib.timeRangeIndicator == 124):
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("sum", coords="time"))

    if \
            (grib.timeRangeIndicator == 125):
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("standard_deviation", coords="time"))

    if \
            (grib.levelType == 'pl'):
        aux_coords_and_dims.append((DimCoord(points=grib.level,  long_name="pressure", units="hPa"), None))

    if \
            (grib.levelType == 'sfc'):

            if (grib._cf_data is not None) and \
            (grib._cf_data.set_height is not None):
                aux_coords_and_dims.append((DimCoord(points=grib._cf_data.set_height,  long_name="height", units="m", attributes={'positive':'up'}), None))
            elif grib.typeOfLevel == 'heightAboveGround': # required for NCAR
                aux_coords_and_dims.append((DimCoord(points=grib.level,  long_name="height", units="m", attributes={'positive':'up'}), None))

    if \
            (grib.levelType == 'ml') and \
            (hasattr(grib, 'pv')):
        aux_coords_and_dims.append((AuxCoord(grib.level, standard_name='model_level_number', units=1, attributes={'positive': 'up'}), None))
        aux_coords_and_dims.append((DimCoord(grib.pv[grib.level], long_name='level_pressure', units='Pa'), None))
        aux_coords_and_dims.append((AuxCoord(grib.pv[grib.numberOfCoordinatesValues//2 + grib.level], long_name='sigma', units=1), None))
        factories.append(Factory(HybridPressureFactory, [{'long_name': 'level_pressure'}, {'long_name': 'sigma'}, Reference('surface_pressure')]))

    if grib._originatingCentre != 'unknown':
        aux_coords_and_dims.append((AuxCoord(points=grib._originatingCentre, long_name='originating_centre', units='no_unit'), None))

    return ConversionMetadata(factories, references, standard_name, long_name,
                              units, attributes, cell_methods,
                              dim_coords_and_dims, aux_coords_and_dims)



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris_grib._load_convert.forecast_period_coord.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from iris.coords import DimCoord

from iris_grib._load_convert import forecast_period_coord


class Test(tests.IrisGribTest):
    def test(self):
        # (indicatorOfUnitOfTimeRange, forecastTime, expected-hours)
        times = [(0, 60, 1),     # minutes
                 (1, 2, 2),      # hours
                 (2, 1, 24),     # days
                 (10, 2, 6),     # 3 hours
                 (11, 3, 18),    # 6 hours
                 (12, 2, 24),    # 12 hours
                 (13, 3600, 1)]  # seconds

        for indicatorOfUnitOfTimeRange, forecastTime, hours in times:
            coord = forecast_period_coord(indicatorOfUnitOfTimeRange,
                                          forecastTime)
            self.assertIsInstance(coord, DimCoord)
            self.assertEqual(coord.standard_name, 'forecast_period')
            self.assertEqual(coord.units, 'hours')
            self.assertEqual(coord.shape, (1,))
            self.assertEqual(coord.points[0], hours)
            self.assertFalse(coord.has_bounds())


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests for `iris_grib._load_convert.product_definition_template_32`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from unittest import mock
import warnings

from iris.coords import AuxCoord, DimCoord
import numpy as np

from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib._load_convert import _MDI as MDI

from iris_grib._load_convert import product_definition_template_32


class Test(tests.IrisGribTest):
    def setUp(self):
        self.patch('warnings.warn')
        self.generating_process_patch = self.patch(
            'iris_grib._load_convert.generating_process')
        self.satellite_common_patch = self.patch(
            'iris_grib._load_convert.satellite_common')
        self.time_coords_patch = self.patch(
            'iris_grib._load_convert.time_coords')
        self.data_cutoff_patch = self.patch(
            'iris_grib._load_convert.data_cutoff')

    def test(self, value=10, factor=1):
        # Prepare the arguments.
        series = mock.sentinel.satelliteSeries
        number = mock.sentinel.satelliteNumber
        instrument = mock.sentinel.instrumentType
        rt_coord = mock.sentinel.observation_time
        section = {'NB': 1,
                   'hoursAfterDataCutoff': None,
                   'minutesAfterDataCutoff': None,
                   'satelliteSeries': series,
                   'satelliteNumber': number,
                   'instrumentType': instrument,
                   'scaleFactorOfCentralWaveNumber': 1,
                   'scaledValueOfCentralWaveNumber': 12,
                   }

        # Call the function.
        metadata = empty_metadata()
        product_definition_template_32(section, metadata, rt_coord)

        # Check that 'satellite_common' was called.
        self.assertEqual(self.satellite_common_patch.call_count, 1)
        # Check that 'generating_process' was called.
        self.assertEqual(self.generating_process_patch.call_count, 1)
        # Check that 'data_cutoff' was called.
        self.assertEqual(self.data_cutoff_patch.call_count, 1)
        # Check that 'time_coords' was called.
        self.assertEqual(self.time_coords_patch.call_count, 1)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris_grib._save_rules.grid_definition_template_4`.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import numpy as np

from iris.coord_systems import GeogCS

from iris_grib._save_rules import grid_definition_template_4
from iris_grib.tests.unit.save_rules import GdtTestMixin


class Test(tests.IrisGribTest, GdtTestMixin):
    def setUp(self):
        GdtTestMixin.setUp(self)

    def test__template_number(self):
        grid_definition_template_4(self.test_cube, self.mock_grib)
        self._check_key('gridDefinitionTemplateNumber', 4)

    def test__shape_of_earth_spherical(self):
        cs = GeogCS(semi_major_axis=1.23)
        test_cube = self._make_test_cube(cs=cs)
        grid_definition_template_4(test_cube, self.mock_grib)
        self._check_key('shapeOfTheEarth', 1)
        self._check_key('scaleFactorOfRadiusOfSphericalEarth', 0)
        self._check_key('scaledValueOfRadiusOfSphericalEarth', 1.23)

    def test__shape_of_earth_flattened(self):
        cs = GeogCS(semi_major_axis=1.456,
                    semi_minor_axis=1.123)
        test_cube = self._make_test_cube(cs=cs)
        grid_definition_template_4(test_cube, self.mock_grib)
        self._check_key('shapeOfTheEarth', 7)
        self._check_key('scaleFactorOfEarthMajorAxis', 0)
        self._check_key('scaledValueOfEarthMajorAxis', 1.456)
        self._check_key('scaleFactorOfEarthMinorAxis', 0)
        self._check_key('scaledValueOfEarthMinorAxis', 1.123)

    def test__grid_shape(self):
        test_cube = self._make_test_cube(x_points=np.arange(13),
                                         y_points=np.arange(6))
        grid_definition_template_4(test_cube, self.mock_grib)
        self._check_key('Ni', 13)
        self._check_key('Nj', 6)

    def test__grid_points(self):
        x_floats = np.array([11.0, 12.0, 167.0])
        # TODO: reduce Y to 2 points, when gribapi nx=ny limitation is gone.
        y_floats = np.array([20.0, 21.0, 22.0])
        test_cube = self._make_test_cube(x_points=x_floats, y_points=y_floats)
        grid_definition_template_4(test_cube, self.mock_grib)
        x_longs = np.array(np.round(1e6 * x_floats), dtype=int)
        y_longs = np.array(np.round(1e6 * y_floats), dtype=int)
        self._check_key("longitudes", x_longs)
        self._check_key("latitudes", y_longs)

    def test__scanmode(self):
        grid_definition_template_4(self.test_cube, self.mock_grib)
        self._check_key('iScansPositively', 1)
        self._check_key('jScansPositively', 1)

    def test__scanmode_reverse(self):
        test_cube = self._make_test_cube(x_points=np.arange(7, 0, -1))
        grid_definition_template_4(test_cube, self.mock_grib)
        self._check_key('iScansPositively', 0)
        self._check_key('jScansPositively', 1)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function
:func:`iris_grib._load_convert.ensemble_identifier`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from unittest import mock
import warnings

from iris.coords import DimCoord

from iris_grib._load_convert import ensemble_identifier


class Test(tests.IrisGribTest):
    def setUp(self):
        module = 'iris_grib._load_convert'
        self.patch('warnings.warn')
        this = '{}.product_definition_template_0'.format(module)

    def _check(self, request_warning):
        section = {'perturbationNumber': 17}
        this = 'iris_grib._load_convert.options'
        with mock.patch(this, warn_on_unsupported=request_warning):
            realization = ensemble_identifier(section)
            expected = DimCoord(section['perturbationNumber'],
                                standard_name='realization',
                                units='no_unit')

            if request_warning:
                warn_msgs = [mcall[1][0] for mcall in warnings.warn.mock_calls]
                expected_msgs = ['type of ensemble', 'number of forecasts']
                for emsg in expected_msgs:
                    matches = [wmsg for wmsg in warn_msgs if emsg in wmsg]
                    self.assertEqual(len(matches), 1)
                    warn_msgs.remove(matches[0])
            else:
                self.assertEqual(len(warnings.warn.mock_calls), 0)

    def test_ens_no_warn(self):
        self._check(False)

    def test_ens_warn(self):
        self._check(True)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Integration tests for grib2 file loading.

These tests load various files from iris-test-data, and compare the cube with a
reference CML file, to catch any unexpected changes over time.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import iris

_RESULTDIR_PREFIX = ("integration", "load_convert", "sample_file_loads")


@tests.skip_data
class TestBasicLoad(tests.IrisGribTest):
    def test_load_rotated(self):
        cubes = iris.load(
            tests.get_data_path(("GRIB", "rotated_uk", "uk_wrongparam.grib1"))
        )
        self.assertCML(cubes, _RESULTDIR_PREFIX + ("rotated.cml",))

    def test_load_time_bound(self):
        cubes = iris.load(
            tests.get_data_path(("GRIB", "time_processed", "time_bound.grib1"))
        )
        self.assertCML(cubes, _RESULTDIR_PREFIX + ("time_bound_grib1.cml",))

    def test_load_time_processed(self):
        cubes = iris.load(
            tests.get_data_path(("GRIB", "time_processed", "time_bound.grib2"))
        )
        self.assertCML(cubes, _RESULTDIR_PREFIX + ("time_bound_grib2.cml",))

    def test_load_3_layer(self):
        cubes = iris.load(
            tests.get_data_path(("GRIB", "3_layer_viz", "3_layer.grib2"))
        )
        cubes = iris.cube.CubeList([cubes[1], cubes[0], cubes[2]])
        self.assertCML(cubes, _RESULTDIR_PREFIX + ("3_layer.cml",))

    def test_load_masked(self):
        gribfile = tests.get_data_path(
            ("GRIB", "missing_values", "missing_values.grib2")
        )
        cubes = iris.load(gribfile)
        self.assertCML(cubes,
                       _RESULTDIR_PREFIX + ("missing_values_grib2.cml",))

    def test_polar_stereo_grib1(self):
        cube = iris.load_cube(
            tests.get_data_path(("GRIB", "polar_stereo", "ST4.2013052210.01h"))
        )
        self.assertCML(cube, _RESULTDIR_PREFIX + ("polar_stereo_grib1.cml",))

    def test_polar_stereo_grib2_grid_definition(self):
        cube = iris.load_cube(
            tests.get_data_path(
                (
                    "GRIB",
                    "polar_stereo",
                    "CMC_glb_TMP_ISBL_1015_ps30km_2013052000_P006.grib2",
                )
            )
        )
        self.assertEqual(cube.shape, (200, 247))
        pxc = cube.coord("projection_x_coordinate")
        self.assertAlmostEqual(pxc.points.max(), 4769905.5125, places=4)
        self.assertAlmostEqual(pxc.points.min(), -2610094.4875, places=4)
        pyc = cube.coord("projection_y_coordinate")
        self.assertAlmostEqual(pyc.points.max(), -216.1459, places=4)
        self.assertAlmostEqual(pyc.points.min(), -5970216.1459, places=4)
        self.assertEqual(pyc.coord_system, pxc.coord_system)
        self.assertEqual(pyc.coord_system.grid_mapping_name, "stereographic")
        self.assertEqual(pyc.coord_system.central_lat, 90.0)
        self.assertEqual(pyc.coord_system.central_lon, 249.0)
        self.assertEqual(pyc.coord_system.false_easting, 0.0)
        self.assertEqual(pyc.coord_system.false_northing, 0.0)
        self.assertEqual(pyc.coord_system.true_scale_lat, 60.0)

    def test_lambert_grib1(self):
        cube = iris.load_cube(
            tests.get_data_path(("GRIB", "lambert", "lambert.grib1"))
        )
        self.assertCML(cube, _RESULTDIR_PREFIX + ("lambert_grib1.cml",))

    def test_lambert_grib2(self):
        cube = iris.load_cube(
            tests.get_data_path(("GRIB", "lambert", "lambert.grib2"))
        )
        self.assertCML(cube, _RESULTDIR_PREFIX + ("lambert_grib2.cml",))

    def test_regular_gg_grib1(self):
        cube = iris.load_cube(
            tests.get_data_path(("GRIB", "gaussian", "regular_gg.grib1"))
        )
        self.assertCML(cube, _RESULTDIR_PREFIX + ("regular_gg_grib1.cml",))

    def test_regular_gg_grib2(self):
        cube = iris.load_cube(
            tests.get_data_path(("GRIB", "gaussian", "regular_gg.grib2"))
        )
        self.assertCML(cube, _RESULTDIR_PREFIX + ("regular_gg_grib2.cml",))

    def test_reduced_ll(self):
        cube = iris.load_cube(
            tests.get_data_path(("GRIB", "reduced", "reduced_ll.grib1"))
        )
        self.assertCML(cube, _RESULTDIR_PREFIX + ("reduced_ll_grib1.cml",))

    def test_reduced_gg(self):
        cube = iris.load_cube(
            tests.get_data_path(("GRIB", "reduced", "reduced_gg.grib2"))
        )
        self.assertCML(cube, _RESULTDIR_PREFIX + ("reduced_gg_grib2.cml",))


@tests.skip_data
class TestIjDirections(tests.IrisGribTest):
    @staticmethod
    def _old_compat_load(name):
        filepath = tests.get_data_path(("GRIB", "ij_directions", name))
        cube = iris.load_cube(filepath)
        return cube

    def test_ij_directions_ipos_jpos(self):
        cubes = self._old_compat_load("ipos_jpos.grib2")
        self.assertCML(cubes, _RESULTDIR_PREFIX + ("ipos_jpos.cml",))

    def test_ij_directions_ipos_jneg(self):
        cubes = self._old_compat_load("ipos_jneg.grib2")
        self.assertCML(cubes, _RESULTDIR_PREFIX + ("ipos_jneg.cml",))

    def test_ij_directions_ineg_jneg(self):
        cubes = self._old_compat_load("ineg_jneg.grib2")
        self.assertCML(cubes, _RESULTDIR_PREFIX + ("ineg_jneg.cml",))

    def test_ij_directions_ineg_jpos(self):
        cubes = self._old_compat_load("ineg_jpos.grib2")
        self.assertCML(cubes, _RESULTDIR_PREFIX + ("ineg_jpos.cml",))


@tests.skip_data
class TestShapeOfEarth(tests.IrisGribTest):
    @staticmethod
    def _old_compat_load(name):
        filepath = tests.get_data_path(("GRIB", "shape_of_earth", name))
        cube = iris.load_cube(filepath)
        return cube

    def test_shape_of_earth_basic(self):
        # pre-defined sphere
        cube = self._old_compat_load("0.grib2")
        self.assertCML(cube, _RESULTDIR_PREFIX + ("earth_shape_0.cml",))

    def test_shape_of_earth_custom_1(self):
        # custom sphere
        cube = self._old_compat_load("1.grib2")
        self.assertCML(cube, _RESULTDIR_PREFIX + ("earth_shape_1.cml",))

    def test_shape_of_earth_IAU65(self):
        # IAU65 oblate sphere
        cube = self._old_compat_load("2.grib2")
        self.assertCML(cube, _RESULTDIR_PREFIX + ("earth_shape_2.cml",))

    def test_shape_of_earth_custom_3(self):
        # custom oblate spheroid (km)
        cube = self._old_compat_load("3.grib2")
        self.assertCML(cube, _RESULTDIR_PREFIX + ("earth_shape_3.cml",))

    def test_shape_of_earth_IAG_GRS80(self):
        # IAG-GRS80 oblate spheroid
        cube = self._old_compat_load("4.grib2")
        self.assertCML(cube, _RESULTDIR_PREFIX + ("earth_shape_4.cml",))

    def test_shape_of_earth_WGS84(self):
        # WGS84
        cube = self._old_compat_load("5.grib2")
        self.assertCML(cube, _RESULTDIR_PREFIX + ("earth_shape_5.cml",))

    def test_shape_of_earth_pre_6(self):
        # pre-defined sphere
        cube = self._old_compat_load("6.grib2")
        self.assertCML(cube, _RESULTDIR_PREFIX + ("earth_shape_6.cml",))

    def test_shape_of_earth_custom_7(self):
        # custom oblate spheroid (m)
        cube = self._old_compat_load("7.grib2")
        self.assertCML(cube, _RESULTDIR_PREFIX + ("earth_shape_7.cml",))

    def test_shape_of_earth_grib1(self):
        # grib1 - same as grib2 shape 6, above
        cube = self._old_compat_load("global.grib1")
        self.assertCML(cube, _RESULTDIR_PREFIX + ("earth_shape_grib1.cml",))


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris_grib._save_rules.set_time_range`

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock
import warnings

from cf_units import Unit
import gribapi

from iris.coords import DimCoord

from iris_grib._save_rules import set_time_range


class Test(tests.IrisGribTest):
    def setUp(self):
        self.coord = DimCoord(0, 'time',
                              units=Unit('hours since epoch',
                                         calendar='standard'))

    def test_no_bounds(self):
        with self.assertRaisesRegex(ValueError, 'Expected time coordinate '
                                    'with two bounds, got 0 bounds'):
            set_time_range(self.coord, mock.sentinel.grib)

    def test_three_bounds(self):
        self.coord.bounds = [0, 1, 2]
        with self.assertRaisesRegex(ValueError, 'Expected time coordinate '
                                    'with two bounds, got 3 bounds'):
            set_time_range(self.coord, mock.sentinel.grib)

    def test_non_scalar(self):
        coord = DimCoord([0, 1], 'time', bounds=[[0, 1], [1, 2]],
                         units=Unit('hours since epoch', calendar='standard'))
        with self.assertRaisesRegex(ValueError, 'Expected length one time '
                                    'coordinate, got 2 points'):
            set_time_range(coord, mock.sentinel.grib)

    @mock.patch.object(gribapi, 'grib_set')
    def test_hours(self, mock_set):
        lower = 10
        upper = 20
        self.coord.bounds = [lower, upper]
        set_time_range(self.coord, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 'indicatorOfUnitForTimeRange', 1)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 'lengthOfTimeRange', upper - lower)

    @mock.patch.object(gribapi, 'grib_set')
    def test_days(self, mock_set):
        lower = 4
        upper = 6
        self.coord.bounds = [lower, upper]
        self.coord.units = Unit('days since epoch', calendar='standard')
        set_time_range(self.coord, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 'indicatorOfUnitForTimeRange', 1)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 'lengthOfTimeRange',
                                 (upper - lower) * 24)

    @mock.patch.object(gribapi, 'grib_set')
    def test_fractional_hours(self, mock_set_long):
        lower = 10.0
        upper = 20.9
        self.coord.bounds = [lower, upper]
        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            set_time_range(self.coord, mock.sentinel.grib)
        self.assertEqual(len(warn), 1)
        msg = 'Truncating floating point lengthOfTimeRange 10\.8?9+ ' \
              'to integer value 10'
        self.assertRegex(str(warn[0].message), msg)
        mock_set_long.assert_any_call(mock.sentinel.grib,
                                      'indicatorOfUnitForTimeRange', 1)
        mock_set_long.assert_any_call(mock.sentinel.grib,
                                      'lengthOfTimeRange', int(upper - lower))


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Temporary module to check for the extended Mercator class in Iris,
which iris-grib requires for its Mercator support.

"""

import distutils.version

import iris


def confirm_extended_mercator_supported():
    # Check that Iris version is at least 2.1, required for 'standard_parallel'
    # support in the Mercator coord-system.
    # This is a temporary fix allowing us to state Iris>=2.0 as a dependency,
    # required for this release because Iris 2.1 is not yet available.
    iris_version = distutils.version.LooseVersion(iris.__version__)
    min_mercator_version = '2.1.0'
    if iris_version < min_mercator_version:
        msg = 'Support for Mercator projections requires Iris version >= {}'
        raise ValueError(msg.format(min_mercator_version))



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests for specific implementation aspects of the grib saver.
Ported here from 'iris.tests.integration.test_grib_save'.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else
import iris_grib.tests as tests

import os
import datetime

import cf_units
import numpy as np

import iris
import iris.cube
import iris.coord_systems
import iris.coords
import iris.exceptions
import iris.util

import gribapi
from iris_grib._load_convert import _MDI as MDI


class TestLoadSave(tests.TestGribMessage):
    def setUp(self):
        self.skip_keys = []

    def test_latlon_forecast_plev(self):
        source_grib = tests.get_data_path(("GRIB", "uk_t", "uk_t.grib2"))
        cubes = iris.load(source_grib)
        with self.temp_filename(suffix=".grib2") as temp_file_path:
            iris.save(cubes, temp_file_path)
            expect_diffs = {
                "totalLength": (4837, 4832),
                "productionStatusOfProcessedData": (0, 255),
                "scaleFactorOfRadiusOfSphericalEarth": (MDI, 0),
                "shapeOfTheEarth": (0, 1),
                "scaledValueOfRadiusOfSphericalEarth": (MDI, 6367470),
                "scaledValueOfEarthMajorAxis": (MDI, 0),
                "scaleFactorOfEarthMajorAxis": (MDI, 0),
                "scaledValueOfEarthMinorAxis": (MDI, 0),
                "scaleFactorOfEarthMinorAxis": (MDI, 0),
                "typeOfGeneratingProcess": (0, 255),
                "generatingProcessIdentifier": (128, 255),
            }
            self.assertGribMessageDifference(
                source_grib,
                temp_file_path,
                expect_diffs,
                self.skip_keys,
                skip_sections=[2],
            )

    def test_rotated_latlon(self):
        source_grib = tests.get_data_path(
            ("GRIB", "rotated_nae_t", "sensible_pole.grib2")
        )
        cubes = iris.load(source_grib)
        with self.temp_filename(suffix=".grib2") as temp_file_path:
            iris.save(cubes, temp_file_path)
            expect_diffs = {
                "totalLength": (648196, 648191),
                "productionStatusOfProcessedData": (0, 255),
                "scaleFactorOfRadiusOfSphericalEarth": (MDI, 0),
                "shapeOfTheEarth": (0, 1),
                "scaledValueOfRadiusOfSphericalEarth": (MDI, 6367470),
                "scaledValueOfEarthMajorAxis": (MDI, 0),
                "scaleFactorOfEarthMajorAxis": (MDI, 0),
                "scaledValueOfEarthMinorAxis": (MDI, 0),
                "scaleFactorOfEarthMinorAxis": (MDI, 0),
                "longitudeOfLastGridPoint": (392109982, 32106370),
                "latitudeOfLastGridPoint": (19419996, 19419285),
                "typeOfGeneratingProcess": (0, 255),
                "generatingProcessIdentifier": (128, 255),
            }
            self.assertGribMessageDifference(
                source_grib,
                temp_file_path,
                expect_diffs,
                self.skip_keys,
                skip_sections=[2],
            )

    def test_time_mean(self):
        # This test for time-mean fields also tests negative forecast time.
        source_grib = tests.get_data_path(
            ("GRIB", "time_processed", "time_bound.grib2")
        )
        cubes = iris.load(source_grib)
        expect_diffs = {
            "totalLength": (21232, 21227),
            "productionStatusOfProcessedData": (0, 255),
            "scaleFactorOfRadiusOfSphericalEarth": (MDI, 0),
            "shapeOfTheEarth": (0, 1),
            "scaledValueOfRadiusOfSphericalEarth": (MDI, 6367470),
            "scaledValueOfEarthMajorAxis": (MDI, 0),
            "scaleFactorOfEarthMajorAxis": (MDI, 0),
            "scaledValueOfEarthMinorAxis": (MDI, 0),
            "scaleFactorOfEarthMinorAxis": (MDI, 0),
            "longitudeOfLastGridPoint": (356249908, 356249809),
            "latitudeOfLastGridPoint": (-89999938, -89999944),
            "typeOfGeneratingProcess": (0, 255),
            "generatingProcessIdentifier": (128, 255),
            "typeOfTimeIncrement": (2, 255),
        }
        self.skip_keys.append("stepType")
        self.skip_keys.append("stepTypeInternal")
        with self.temp_filename(suffix=".grib2") as temp_file_path:
            iris.save(cubes, temp_file_path)
            self.assertGribMessageDifference(
                source_grib,
                temp_file_path,
                expect_diffs,
                self.skip_keys,
                skip_sections=[2],
            )


class TestCubeSave(tests.IrisGribTest):
    # save fabricated cubes

    def _load_basic(self):
        path = tests.get_data_path(("GRIB", "uk_t", "uk_t.grib2"))
        return iris.load_cube(path)

    def test_params(self):
        # TODO
        pass

    def test_originating_centre(self):
        # TODO
        pass

    def test_irregular(self):
        cube = self._load_basic()
        lat_coord = cube.coord("latitude")
        cube.remove_coord("latitude")

        new_lats = np.append(
            lat_coord.points[:-1], lat_coord.points[0]
        )  # Irregular
        cube.add_aux_coord(
            iris.coords.AuxCoord(
                new_lats,
                "latitude",
                units="degrees",
                coord_system=lat_coord.coord_system,
            ),
            0,
        )

        saved_grib = iris.util.create_temp_filename(suffix=".grib2")
        self.assertRaises(
            iris.exceptions.TranslationError, iris.save, cube, saved_grib
        )
        os.remove(saved_grib)

    def test_non_latlon(self):
        cube = self._load_basic()
        cube.coord(dimensions=[0]).coord_system = None
        saved_grib = iris.util.create_temp_filename(suffix=".grib2")
        self.assertRaises(
            iris.exceptions.TranslationError, iris.save, cube, saved_grib
        )
        os.remove(saved_grib)

    def test_forecast_period(self):
        # unhandled unit
        cube = self._load_basic()
        cube.coord("forecast_period").units = cf_units.Unit("years")
        saved_grib = iris.util.create_temp_filename(suffix=".grib2")
        self.assertRaises(
            iris.exceptions.TranslationError, iris.save, cube, saved_grib
        )
        os.remove(saved_grib)

    def test_unhandled_vertical(self):
        # unhandled level type
        cube = self._load_basic()
        # Adjust the 'pressure' coord to make it into an "unrecognised Z coord"
        p_coord = cube.coord("pressure")
        p_coord.rename("not the messiah")
        p_coord.units = "K"
        p_coord.attributes["positive"] = "up"
        saved_grib = iris.util.create_temp_filename(suffix=".grib2")
        with self.assertRaises(iris.exceptions.TranslationError):
            iris.save(cube, saved_grib)
        os.remove(saved_grib)

    def test_scalar_int32_pressure(self):
        # Make sure we can save a scalar int32 coordinate with unit conversion.
        cube = self._load_basic()
        cube.coord("pressure").points = np.array([200], dtype=np.int32)
        cube.coord("pressure").units = "hPa"
        with self.temp_filename(".grib2") as testfile:
            iris.save(cube, testfile)

    def test_bounded_level(self):
        cube = iris.load_cube(
            tests.get_data_path(("GRIB", "uk_t", "uk_t.grib2"))
        )
        with self.temp_filename(".grib2") as testfile:
            iris.save(cube, testfile)
            with open(testfile, "rb") as saved_file:
                g = gribapi.grib_new_from_file(saved_file)
                self.assertEqual(
                    gribapi.grib_get_double(
                        g, "scaledValueOfFirstFixedSurface"
                    ),
                    0.0,
                )
                self.assertEqual(
                    gribapi.grib_get_double(
                        g, "scaledValueOfSecondFixedSurface"
                    ),
                    2147483647.0,
                )


class TestHandmade(tests.IrisGribTest):
    def _lat_lon_cube_no_time(self):
        """
        Returns a cube with a latitude and longitude suitable for testing
        saving to GRIB.

        """
        cube = iris.cube.Cube(np.arange(12, dtype=np.int32).reshape((3, 4)))
        cs = iris.coord_systems.GeogCS(6371229)
        cube.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(4) * 90 + -180,
                "longitude",
                units="degrees",
                coord_system=cs,
            ),
            1,
        )
        cube.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(3) * 45 + -90,
                "latitude",
                units="degrees",
                coord_system=cs,
            ),
            0,
        )

        return cube

    def _cube_time_no_forecast(self):
        cube = self._lat_lon_cube_no_time()
        unit = cf_units.Unit(
            "hours since epoch", calendar=cf_units.CALENDAR_GREGORIAN
        )
        dt = datetime.datetime(2010, 12, 31, 12, 0)
        cube.add_aux_coord(
            iris.coords.AuxCoord(
                np.array([unit.date2num(dt)], dtype=np.float64),
                "time",
                units=unit,
            )
        )
        return cube

    def _cube_with_forecast(self):
        cube = self._cube_time_no_forecast()
        cube.add_aux_coord(
            iris.coords.AuxCoord(
                np.array([6], dtype=np.int32), "forecast_period", units="hours"
            )
        )
        return cube

    def _cube_with_pressure(self):
        cube = self._cube_with_forecast()
        cube.add_aux_coord(
            iris.coords.DimCoord(np.int32(10), "air_pressure", units="Pa")
        )
        return cube

    def _cube_with_time_bounds(self):
        cube = self._cube_with_pressure()
        cube.coord("time").bounds = np.array([[0, 100]])
        return cube

    def test_no_time_cube(self):
        cube = self._lat_lon_cube_no_time()
        saved_grib = iris.util.create_temp_filename(suffix=".grib2")
        self.assertRaises(
            iris.exceptions.TranslationError, iris.save, cube, saved_grib
        )
        os.remove(saved_grib)

    def test_cube_with_time_bounds(self):
        cube = self._cube_with_time_bounds()
        saved_grib = iris.util.create_temp_filename(suffix=".grib2")
        self.assertRaises(
            iris.exceptions.TranslationError, iris.save, cube, saved_grib
        )
        os.remove(saved_grib)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Integration tests to confirm data is loaded correctly.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import numpy.ma as ma

from iris import load_cube


class TestImport(tests.IrisGribTest):
    def test_gdt1(self):
        path = tests.get_data_path(
            ("GRIB", "rotated_nae_t", "sensible_pole.grib2")
        )
        cube = load_cube(path)
        self.assertCMLApproxData(cube)

    def test_gdt90_with_bitmap(self):
        path = tests.get_data_path(("GRIB", "umukv", "ukv_chan9.grib2"))
        cube = load_cube(path)
        # Pay particular attention to the orientation.
        self.assertIsNot(cube.data[0, 0], ma.masked)
        self.assertIs(cube.data[-1, 0], ma.masked)
        self.assertIs(cube.data[0, -1], ma.masked)
        self.assertIs(cube.data[-1, -1], ma.masked)
        x = cube.coord("projection_x_coordinate").points
        y = cube.coord("projection_y_coordinate").points
        self.assertGreater(x[0], x[-1])  # Decreasing X coordinate
        self.assertLess(y[0], y[-1])  # Increasing Y coordinate
        # Check everything else.
        self.assertCMLApproxData(cube)


class TestGDT30(tests.IrisGribTest):
    def test_lambert(self):
        path = tests.get_data_path(("GRIB", "lambert", "lambert.grib2"))
        cube = load_cube(path)
        self.assertCMLApproxData(cube)


class TestGDT40(tests.IrisGribTest):
    def test_regular(self):
        path = tests.get_data_path(("GRIB", "gaussian", "regular_gg.grib2"))
        cube = load_cube(path)
        self.assertCMLApproxData(cube)

    def test_reduced(self):
        path = tests.get_data_path(("GRIB", "reduced", "reduced_gg.grib2"))
        cube = load_cube(path)
        self.assertCMLApproxData(cube)


class TestDRT3(tests.IrisGribTest):
    def test_grid_complex_spatial_differencing(self):
        path = tests.get_data_path(
            ("GRIB", "missing_values", "missing_values.grib2")
        )
        cube = load_cube(path)
        self.assertCMLApproxData(cube)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris_grib._load_convert.reference_time_coord.

Reference Code Table 1.2.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from datetime import datetime

from cf_units import CALENDAR_GREGORIAN, Unit

from iris.coords import DimCoord
from iris.exceptions import TranslationError

from iris_grib._load_convert import reference_time_coord


class Test(tests.IrisGribTest):
    def setUp(self):
        self.section = {'year': 2007,
                        'month': 1,
                        'day': 15,
                        'hour': 0,
                        'minute': 3,
                        'second': 0}
        self.unit = Unit('hours since epoch', calendar=CALENDAR_GREGORIAN)
        dt = datetime(self.section['year'], self.section['month'],
                      self.section['day'], self.section['hour'],
                      self.section['minute'], self.section['second'])
        self.point = self.unit.date2num(dt)

    def _check(self, section, standard_name=None):
        expected = DimCoord(self.point, standard_name=standard_name,
                            units=self.unit)
        # The call being tested.
        coord = reference_time_coord(section)
        self.assertEqual(coord, expected)

    def test_start_of_forecast__0(self):
        section = deepcopy(self.section)
        section['significanceOfReferenceTime'] = 0
        self._check(section, 'forecast_reference_time')

    def test_start_of_forecast__1(self):
        section = deepcopy(self.section)
        section['significanceOfReferenceTime'] = 1
        self._check(section, 'forecast_reference_time')

    def test_observation_time(self):
        section = deepcopy(self.section)
        section['significanceOfReferenceTime'] = 3
        self._check(section, 'time')

    def test_unknown_significance(self):
        section = deepcopy(self.section)
        section['significanceOfReferenceTime'] = 5
        emsg = 'unsupported significance'
        with self.assertRaisesRegex(TranslationError, emsg):
            self._check(section)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Provides testing capabilities for installed copies of iris-grib.

"""

# Because this file is imported by setup.py, there may be additional runtime
# imports later in the file.
import multiprocessing
import os


# NOTE: Do not inherit from object as distutils does not like it.
class TestRunner:
    """Run the iris-grib tests under nose and multiprocessor for performance"""

    description = (
        "Run tests under nose and multiprocessor for performance. "
        "Specifying one or more test flags will run *only* those "
        "tests."
    )
    user_options = [
        (
            "no-data",
            "n",
            "Override the paths to the data repositories so it "
            "appears to the tests that it does not exist.",
        ),
        ("stop", "x", "Stop running tests after the first error or failure."),
        ("unit-tests", "s", "Run the limited subset of unit tests."),
        ("integration-tests", "i", "Run the integration tests."),
        ("default-tests", "d", "Run the default tests."),
        (
            "coding-tests",
            "c",
            "Run the coding standards tests. (These are a "
            "subset of the default tests.)",
        ),
        (
            "num-processors=",
            "p",
            "The number of processors used for running " "the tests.",
        ),
        ("create-missing", "m", "Create missing test result files."),
    ]
    boolean_options = [
        "no-data",
        "unit-tests",
        "stop",
        "default-tests",
        "integration-tests",
        "create-missing",
    ]

    def initialize_options(self):
        self.no_data = False
        self.stop = False
        self.unit_tests = False
        self.default_tests = False
        self.integration_tests = False
        self.num_processors = None
        self.create_missing = False

    def finalize_options(self):
        # These enviroment variables will be propagated to all the
        # processes that nose.run creates.
        if self.no_data:
            print("Running tests in no-data mode...")
            import iris.config

            iris.config.TEST_DATA_DIR = None
        if self.create_missing:
            os.environ["IRIS_TEST_CREATE_MISSING"] = "true"

        tests = []
        if self.unit_tests:
            tests.append("unit")
        if self.default_tests:
            tests.append("default")
        if self.integration_tests:
            tests.append("integration")
        if not tests:
            tests.append("default")
        print("Running test suite(s): {}".format(", ".join(tests)))
        if self.stop:
            print("Stopping tests after the first error or failure")
        if self.num_processors is None:
            # Choose a magic number that works reasonably well for the default
            # number of processes.
            self.num_processors = (multiprocessing.cpu_count() + 1) // 4 + 1
        else:
            self.num_processors = int(self.num_processors)

    def run(self):
        import nose

        if hasattr(self, "distribution") and self.distribution.tests_require:
            self.distribution.fetch_build_eggs(self.distribution.tests_require)

        tests = []
        if self.unit_tests:
            tests.append("iris_grib.tests.unit")
        if self.default_tests:
            tests.append("iris_grib.tests")
        if self.integration_tests:
            tests.append("iris_grib.tests.integration")

        if not tests:
            tests.append("iris_grib.tests")

        regexp_pat = r"--match=^([Tt]est(?![Mm]ixin)|[Ss]ystem)"

        n_processors = max(self.num_processors, 1)

        args = [
            "",
            None,
            "--processes=%s" % n_processors,
            "--verbosity=2",
            regexp_pat,
            "--process-timeout=180",
        ]

        if self.stop:
            args.append("--stop")

        result = True
        for test in tests:
            args[1] = test
            print()
            print(
                "Running test discovery on %s with %s processors."
                % (test, n_processors)
            )
            # run the tests at module level i.e. my_module.tests
            # - test must start with test/Test and must not contain the
            #   word Mixin.
            result &= nose.run(argv=args)
        if result is False:
            exit(1)



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :mod:`iris_grib.grib_save_rules` module."""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock

import numpy as np

from iris.coords import DimCoord
from iris.coord_systems import GeogCS
from iris.cube import Cube
from iris.fileformats.pp import EARTH_RADIUS as PP_DEFAULT_EARTH_RADIUS


class GdtTestMixin:
    """Some handy common test capabilities for grib grid-definition tests."""
    TARGET_MODULE = 'iris_grib._save_rules'

    def setUp(self):
        # Patch the gribapi of the tested module.
        self.mock_gribapi = self.patch(self.TARGET_MODULE + '.gribapi')

        # Fix the mock gribapi to record key assignments.
        def grib_set_trap(grib, name, value):
            # Record a key setting on the mock passed as the 'grib message id'.
            grib.keys[name] = value

        self.mock_gribapi.grib_set = grib_set_trap
        self.mock_gribapi.grib_set_long = grib_set_trap
        self.mock_gribapi.grib_set_float = grib_set_trap
        self.mock_gribapi.grib_set_double = grib_set_trap
        self.mock_gribapi.grib_set_long_array = grib_set_trap
        self.mock_gribapi.grib_set_array = grib_set_trap

        # Create a mock 'grib message id', with a 'keys' dict for settings.
        self.mock_grib = mock.Mock(keys={})

        # Initialise the test cube and its coords to something barely usable.
        self.test_cube = self._make_test_cube()

    def _default_coord_system(self):
        return GeogCS(PP_DEFAULT_EARTH_RADIUS)

    def _default_x_points(self):
        # Define simple, regular coordinate points.
        return [1.0, 2.0, 3.0]

    def _default_y_points(self):
        return [7.0, 8.0]  # N.B. is_regular will *fail* on length-1 coords.

    def _make_test_cube(self, cs=None, x_points=None, y_points=None,
                        coord_units='degrees'):
        # Create a cube with given properties, or minimal defaults.
        if cs is None:
            cs = self._default_coord_system()
        if x_points is None:
            x_points = self._default_x_points()
        if y_points is None:
            y_points = self._default_y_points()

        x_coord = DimCoord(x_points, long_name='longitude',
                           units=coord_units, coord_system=cs)
        y_coord = DimCoord(y_points, long_name='latitude',
                           units=coord_units, coord_system=cs)
        test_cube = Cube(np.zeros((len(y_points), len(x_points))))
        test_cube.add_dim_coord(y_coord, 0)
        test_cube.add_dim_coord(x_coord, 1)
        return test_cube

    def _check_key(self, name, value):
        # Test that a specific grib key assignment occurred.
        msg_fmt = 'Expected grib setting "{}" = {}, got {}'
        found = self.mock_grib.keys.get(name)
        if found is None:
            self.assertEqual(0, 1, msg_fmt.format(name, value, '((UNSET))'))
        else:
            self.assertArrayEqual(found, value,
                                  msg_fmt.format(name, value, found))



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests for function
:func:`iris_grib._load_convert.statistical_cell_method`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from iris.coords import CellMethod
from iris.exceptions import TranslationError

from iris_grib._load_convert import statistical_cell_method


class Test(tests.IrisGribTest):
    def setUp(self):
        self.section = {}
        self.section['productDefinitionTemplateNumber'] = 8
        self.section['numberOfTimeRange'] = 1
        self.section['typeOfStatisticalProcessing'] = 0
        self.section['typeOfTimeIncrement'] = 2
        self.section['timeIncrement'] = 0

    def expected_cell_method(self,
                             coords=('time',), method='mean', intervals=None):
        keys = dict(coords=coords, method=method, intervals=intervals)
        cell_method = CellMethod(**keys)
        return cell_method

    def test_basic(self):
        cell_method = statistical_cell_method(self.section)
        self.assertEqual(cell_method, self.expected_cell_method())

    def test_intervals(self):
        self.section['timeIncrement'] = 3
        self.section['indicatorOfUnitForTimeIncrement'] = 1
        cell_method = statistical_cell_method(self.section)
        self.assertEqual(cell_method,
                         self.expected_cell_method(intervals=('3 hours',)))

    def test_increment_missing(self):
        self.section['timeIncrement'] = 2 ** 32 - 1
        self.section['indicatorOfUnitForTimeIncrement'] = 255
        cell_method = statistical_cell_method(self.section)
        self.assertEqual(cell_method, self.expected_cell_method())

    def test_different_statistic(self):
        self.section['typeOfStatisticalProcessing'] = 6
        cell_method = statistical_cell_method(self.section)
        self.assertEqual(
            cell_method,
            self.expected_cell_method(method='standard_deviation'))

    def test_fail_bad_ranges(self):
        self.section['numberOfTimeRange'] = 0
        with self.assertRaisesRegex(TranslationError,
                                    'aggregation over "0 time ranges"'):
            statistical_cell_method(self.section)

    def test_fail_multiple_ranges(self):
        self.section['numberOfTimeRange'] = 2
        with self.assertRaisesRegex(TranslationError,
                                    'multiple time ranges \[2\]'):
            statistical_cell_method(self.section)

    def test_fail_unknown_statistic(self):
        self.section['typeOfStatisticalProcessing'] = 17
        with self.assertRaisesRegex(
                TranslationError,
                'contains an unsupported statistical process type \[17\]'):
            statistical_cell_method(self.section)

    def test_fail_bad_increment_type(self):
        self.section['typeOfTimeIncrement'] = 7
        with self.assertRaisesRegex(
                TranslationError,
                'time-increment type \[7\] is not supported'):
            statistical_cell_method(self.section)

    def test_pdt_9(self):
        # Should behave the same as PDT 4.8.
        self.section['productDefinitionTemplateNumber'] = 9
        cell_method = statistical_cell_method(self.section)
        self.assertEqual(cell_method, self.expected_cell_method())

    def test_pdt_10(self):
        # Should behave the same as PDT 4.8.
        self.section['productDefinitionTemplateNumber'] = 10
        cell_method = statistical_cell_method(self.section)
        self.assertEqual(cell_method, self.expected_cell_method())

    def test_pdt_11(self):
        # Should behave the same as PDT 4.8.
        self.section['productDefinitionTemplateNumber'] = 11
        cell_method = statistical_cell_method(self.section)
        self.assertEqual(cell_method, self.expected_cell_method())

    def test_pdt_15(self):
        # Encoded slightly differently to PDT 4.8.
        self.section['productDefinitionTemplateNumber'] = 15
        test_code = self.section['typeOfStatisticalProcessing']
        del self.section['typeOfStatisticalProcessing']
        self.section['statisticalProcess'] = test_code
        cell_method = statistical_cell_method(self.section)
        self.assertEqual(cell_method, self.expected_cell_method())

    def test_fail_unsupported_pdt(self):
        # Rejects PDTs other than the ones tested above.
        self.section['productDefinitionTemplateNumber'] = 101
        msg = "can't get statistical method for unsupported pdt : 4.101"
        with self.assertRaisesRegex(ValueError, msg):
            statistical_cell_method(self.section)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests for `iris_grib._load_convert.product_definition_template_31`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from unittest import mock
import warnings

from iris.coords import AuxCoord
import numpy as np

from iris_grib.tests.unit.load_convert import empty_metadata

from iris_grib._load_convert import product_definition_template_31


def section_4():
    # Also needed for test_product_definition_section.py.
    series = mock.sentinel.satelliteSeries
    number = mock.sentinel.satelliteNumber
    instrument = mock.sentinel.instrumentType
    return {'NB': 1,
            'satelliteSeries': series,
            'satelliteNumber': number,
            'instrumentType': instrument,
            'scaleFactorOfCentralWaveNumber': 1,
            'scaledValueOfCentralWaveNumber': 12}


class Test(tests.IrisGribTest):
    def setUp(self):
        self.patch('warnings.warn')
        self.satellite_common_patch = self.patch(
            'iris_grib._load_convert.satellite_common')
        self.generating_process_patch = self.patch(
            'iris_grib._load_convert.generating_process')

    def test(self):
        # Prepare the arguments.
        rt_coord = mock.sentinel.observation_time
        section = section_4()

        # Call the function.
        metadata = empty_metadata()
        product_definition_template_31(section, metadata, rt_coord)

        # Check that 'satellite_common' was called.
        self.assertEqual(self.satellite_common_patch.call_count, 1)
        # Check that 'generating_process' was called.
        self.assertEqual(self.generating_process_patch.call_count, 1)
        # Check that the scalar time coord was added in.
        self.assertIn((rt_coord, None), metadata['aux_coords_and_dims'])


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris_grib._load_convert.vertical_coords`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from unittest import mock

from iris.coords import DimCoord
from iris.exceptions import TranslationError

from iris_grib._load_convert import vertical_coords
from iris_grib._load_convert import \
    _TYPE_OF_FIXED_SURFACE_MISSING as MISSING_SURFACE, \
    _MDI as MISSING_LEVEL


class Test(tests.IrisGribTest):
    def setUp(self):
        self.metadata = {'factories': [], 'references': [],
                         'standard_name': None,
                         'long_name': None, 'units': None, 'attributes': {},
                         'cell_methods': [], 'dim_coords_and_dims': [],
                         'aux_coords_and_dims': []}

    def test_hybrid_factories(self):
        def func(section, metadata):
            return metadata['factories'].append(factory)

        metadata = deepcopy(self.metadata)
        section = {'NV': 1}
        this = 'iris_grib._load_convert.hybrid_factories'
        factory = mock.sentinel.factory
        with mock.patch(this, side_effect=func) as hybrid_factories:
            vertical_coords(section, metadata)
            self.assertTrue(hybrid_factories.called)
            self.assertEqual(metadata['factories'], [factory])

    def test_no_first_fixed_surface(self):
        metadata = deepcopy(self.metadata)
        section = {'NV': 0,
                   'typeOfFirstFixedSurface': MISSING_SURFACE,
                   'scaledValueOfFirstFixedSurface': MISSING_LEVEL}
        vertical_coords(section, metadata)
        self.assertEqual(metadata, self.metadata)

    def _check(self, value, msg):
        this = 'iris_grib._load_convert.options'
        with mock.patch('warnings.warn') as warn:
            with mock.patch(this) as options:
                for request_warning in [False, True]:
                    options.warn_on_unsupported = request_warning
                    metadata = deepcopy(self.metadata)
                    section = {'NV': 0,
                               'typeOfFirstFixedSurface': 0,
                               'scaledValueOfFirstFixedSurface': value}
                    # The call being tested.
                    vertical_coords(section, metadata)
                    self.assertEqual(metadata, self.metadata)
                    if request_warning:
                        self.assertEqual(len(warn.mock_calls), 1)
                        args, _ = warn.call_args
                        self.assertIn(msg, args[0])
                    else:
                        self.assertEqual(len(warn.mock_calls), 0)

    def test_unknown_first_fixed_surface_with_missing_scaled_value(self):
        self._check(MISSING_LEVEL, 'surface with missing scaled value')

    def test_unknown_first_fixed_surface_with_scaled_value(self):
        self._check(0, 'surface with scaled value')

    def test_pressure_with_no_second_fixed_surface(self):
        metadata = deepcopy(self.metadata)
        section = {'NV': 0,
                   'typeOfFirstFixedSurface': 100,  # pressure / Pa
                   'scaledValueOfFirstFixedSurface': 10,
                   'scaleFactorOfFirstFixedSurface': 1,
                   'typeOfSecondFixedSurface': MISSING_SURFACE}
        vertical_coords(section, metadata)
        coord = DimCoord(1.0, long_name='pressure', units='Pa')
        expected = deepcopy(self.metadata)
        expected['aux_coords_and_dims'].append((coord, None))
        self.assertEqual(metadata, expected)

    def test_height_with_no_second_fixed_surface(self):
        metadata = deepcopy(self.metadata)
        section = {'NV': 0,
                   'typeOfFirstFixedSurface': 103,  # height / m
                   'scaledValueOfFirstFixedSurface': 100,
                   'scaleFactorOfFirstFixedSurface': 2,
                   'typeOfSecondFixedSurface': MISSING_SURFACE}
        vertical_coords(section, metadata)
        coord = DimCoord(1.0, long_name='height', units='m')
        expected = deepcopy(self.metadata)
        expected['aux_coords_and_dims'].append((coord, None))
        self.assertEqual(metadata, expected)

    def test_different_fixed_surfaces(self):
        section = {'NV': 0,
                   'typeOfFirstFixedSurface': 100,
                   'scaledValueOfFirstFixedSurface': None,
                   'scaleFactorOfFirstFixedSurface': None,
                   'typeOfSecondFixedSurface': 0}
        emsg = 'different types of first and second fixed surface'
        with self.assertRaisesRegex(TranslationError, emsg):
            vertical_coords(section, None)

    def test_same_fixed_surfaces_missing_second_scaled_value(self):
        section = {'NV': 0,
                   'typeOfFirstFixedSurface': 100,
                   'scaledValueOfFirstFixedSurface': None,
                   'scaleFactorOfFirstFixedSurface': None,
                   'typeOfSecondFixedSurface': 100,
                   'scaledValueOfSecondFixedSurface': MISSING_LEVEL}
        emsg = 'missing scaled value of second fixed surface'
        with self.assertRaisesRegex(TranslationError, emsg):
            vertical_coords(section, None)

    def test_pressure_with_second_fixed_surface(self):
        metadata = deepcopy(self.metadata)
        section = {'NV': 0,
                   'typeOfFirstFixedSurface': 100,
                   'scaledValueOfFirstFixedSurface': 10,
                   'scaleFactorOfFirstFixedSurface': 1,
                   'typeOfSecondFixedSurface': 100,
                   'scaledValueOfSecondFixedSurface': 30,
                   'scaleFactorOfSecondFixedSurface': 1}
        vertical_coords(section, metadata)
        coord = DimCoord(2.0, long_name='pressure', units='Pa',
                         bounds=[1.0, 3.0])
        expected = deepcopy(self.metadata)
        expected['aux_coords_and_dims'].append((coord, None))
        self.assertEqual(metadata, expected)

    def test_height_with_second_fixed_surface(self):
        metadata = deepcopy(self.metadata)
        section = {'NV': 0,
                   'typeOfFirstFixedSurface': 103,
                   'scaledValueOfFirstFixedSurface': 1000,
                   'scaleFactorOfFirstFixedSurface': 2,
                   'typeOfSecondFixedSurface': 103,
                   'scaledValueOfSecondFixedSurface': 3000,
                   'scaleFactorOfSecondFixedSurface': 2}
        vertical_coords(section, metadata)
        coord = DimCoord(20.0, long_name='height', units='m',
                         bounds=[10.0, 30.0])
        expected = deepcopy(self.metadata)
        expected['aux_coords_and_dims'].append((coord, None))
        self.assertEqual(metadata, expected)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Integration test for loading hybrid height data.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests


from iris import load_cube, load
from iris.aux_factory import HybridHeightFactory, HybridPressureFactory


@tests.skip_grib_data
class TestHybridHeight(tests.IrisGribTest):
    def setUp(self):
        filepath = self.get_testdata_path('faked_sample_hh_grib_data.grib2')
        self.testdata_cube = load_cube(filepath, 'air_temperature')

    def test_load_hybrid_height(self):
        # Check that it loads right, and creates a factory.
        self.assertIsInstance(self.testdata_cube.aux_factories[0],
                              HybridHeightFactory)

    def test_hybrid_height_coords_values(self):
        cube = self.testdata_cube

        # check actual model level values.
        self.assertArrayEqual(cube.coord('model_level_number').points,
                              [1, 11, 21])

        # check sigma values correctly loaded from indices 1, 11, 21.
        self.assertArrayAllClose(cube.coord('sigma').points,
                                 [0.998, 0.894, 0.667],
                                 atol=0.001)

        # check height values too.
        self.assertArrayAllClose(cube.coord('level_height').points,
                                 [20., 953.3,  3220.],
                                 atol=0.5)


@tests.skip_grib_data
class TestHybridPressure(tests.IrisGribTest):
    def setUp(self):
        filepath = self.get_testdata_path('faked_sample_hp_grib_data.grib2')
        self.testdata_cube = load_cube(filepath, 'air_temperature')

    def test_load_hybrid_pressure(self):
        # Check that it loads right, and creates a factory.
        self.assertIsInstance(self.testdata_cube.aux_factories[0],
                              HybridPressureFactory)

    def test_hybrid_pressure_coords_values(self):
        cube = self.testdata_cube

        # Check existence, and some values, of the loaded coefficients.
        self.assertArrayEqual(cube.coord('model_level_number').points,
                              [1, 51, 91])
        self.assertArrayAllClose(cube.coord('sigma').points,
                                 [0., 0.045, 1.], atol=0.001)
        self.assertArrayAllClose(cube.coord('level_pressure').points,
                                 [2.00004, 18716.9688, 0.], rtol=0.0001)
        self.assertArrayAllClose(
            cube.coord('surface_air_pressure')[:2, :3].points,
            [[103493.8, 103493.8, 103493.8],
             [103401.0, 103407.4, 103412.2]], atol=0.1)

        # Also check a few values from the derived coord.
        self.assertArrayAllClose(
            cube.coord('air_pressure')[:, :3, 0].points,
            [[2., 2., 2.],
             [23389.3, 23385.1, 23379.9],
             [103493.8, 103401.0, 103285.8]], atol=0.1)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Provides testing capabilities and customisations specific to iris-grib.

This imports iris.tests, which requires to be done before anything else for
plot control reasons : see documentation there.

"""

import iris.tests

import inspect
import os
import os.path
import unittest

import numpy as np

try:
    from iris.tests import IrisTest_nometa as IrisTest
except ImportError:
    from iris.tests import IrisTest

from iris.tests import main, skip_data, get_data_path

from iris_grib.message import GribMessage


#: Basepath for iris-grib test results.
_RESULT_PATH = os.path.join(os.path.dirname(__file__), 'results')

#: Basepath for iris-grib loadable test files.
_TESTDATA_PATH = os.path.join(os.path.dirname(__file__), 'testdata')

override = os.environ.get("GRIB_TEST_DATA_PATH")
if override:
    if os.path.isdir(os.path.expanduser(override)):
        _TESTDATA_PATH = os.path.abspath(override)


def skip_grib_data(fn):
    """
    Decorator to choose whether to run tests, based on the availability of
    external data.

    Example usage:
        @skip_data
        class MyDataTests(tests.IrisGribTest):
            ...

    """
    dpath = _TESTDATA_PATH
    evar = "GRIB_TEST_NO_DATA"
    no_data = not os.path.isdir(dpath) or os.environ.get(evar)
    reason = "Test(s) require missing external GRIB test data."

    skip = unittest.skipIf(
        condition=no_data, reason=reason
    )

    return skip(fn)


class IrisGribTest(IrisTest):
    # A specialised version of an IrisTest that implements the correct
    # automatic paths for test results in iris-grib.

    @staticmethod
    def get_result_path(relative_path):
        """
        Returns the absolute path to a result file when given the relative path
        as a string, or sequence of strings.

        """
        if not isinstance(relative_path, str):
            relative_path = os.path.join(*relative_path)
        return os.path.abspath(os.path.join(_RESULT_PATH, relative_path))

    def result_path(self, basename=None, ext=''):
        """
        Return the full path to a test result, generated from the \
        calling file, class and, optionally, method.

        Optional kwargs :

            * basename    - File basename. If omitted, this is \
                            generated from the calling method.
            * ext         - Appended file extension.

        """
        if ext and not ext.startswith('.'):
            ext = '.' + ext

        # Generate the folder name from the calling file name.
        path = os.path.abspath(inspect.getfile(self.__class__))
        path = os.path.splitext(path)[0]
        sub_path = path.rsplit('iris_grib', 1)[1].split('tests', 1)[1][1:]

        # Generate the file name from the calling function name?
        if basename is None:
            stack = inspect.stack()
            for frame in stack[1:]:
                if 'test_' in frame[3]:
                    basename = frame[3].replace('test_', '')
                    break
        filename = basename + ext

        result = os.path.join(self.get_result_path(''),
                              sub_path.replace('test_', ''),
                              self.__class__.__name__.replace('Test_', ''),
                              filename)
        return result

    @staticmethod
    def get_testdata_path(relative_path):
        """
        Returns the absolute path to a loadable test data file, when given the
        relative path as a string, or sequence of strings.

        """
        if not isinstance(relative_path, str):
            relative_path = os.path.join(*relative_path)
        return os.path.abspath(os.path.join(_TESTDATA_PATH, relative_path))


class TestGribMessage(IrisGribTest):
    def assertGribMessageContents(self, filename, contents):
        """
        Evaluate whether all messages in a GRIB2 file contain the provided
        contents.

        * filename (string)
            The path on disk of an existing GRIB file

        * contents
            An iterable of GRIB message keys and expected values.

        """
        messages = GribMessage.messages_from_filename(filename)
        for message in messages:
            for element in contents:
                section, key, val = element
                self.assertEqual(message.sections[section][key], val)

    def assertGribMessageDifference(
        self, filename1, filename2, diffs, skip_keys=(), skip_sections=()
    ):
        """
        Evaluate that the two messages only differ in the ways specified.

        * filename[0|1] (string)
            The path on disk of existing GRIB files

        * diffs
            An dictionary of GRIB message keys and expected diff values:
            {key: (m1val, m2val),...} .

        * skip_keys
            An iterable of key names to ignore during comparison.

        * skip_sections
            An iterable of section numbers to ignore during comparison.

        """
        messages1 = list(GribMessage.messages_from_filename(filename1))
        messages2 = list(GribMessage.messages_from_filename(filename2))
        self.assertEqual(len(messages1), len(messages2))
        for m1, m2 in zip(messages1, messages2):
            m1_sect = set(m1.sections.keys())
            m2_sect = set(m2.sections.keys())

            for missing_section in m1_sect ^ m2_sect:
                what = (
                    "introduced" if missing_section in m1_sect else "removed"
                )
                # Assert that an introduced section is in the diffs.
                self.assertIn(
                    missing_section,
                    skip_sections,
                    msg="Section {} {}".format(missing_section, what),
                )

            for section in m1_sect & m2_sect:
                # For each section, check that the differences are
                # known diffs.
                m1_keys = set(m1.sections[section]._keys)
                m2_keys = set(m2.sections[section]._keys)

                difference = m1_keys ^ m2_keys
                unexpected_differences = difference - set(skip_keys)
                if unexpected_differences:
                    self.fail(
                        "There were keys in section {} which \n"
                        "weren't in both messages and which weren't "
                        "skipped.\n{}"
                        "".format(section, ", ".join(unexpected_differences))
                    )

                keys_to_compare = m1_keys & m2_keys - set(skip_keys)

                for key in keys_to_compare:
                    m1_value = m1.sections[section][key]
                    m2_value = m2.sections[section][key]
                    msg = "{} {} != {}"
                    if key not in diffs:
                        # We have a key which we expect to be the same for
                        # both messages.
                        if isinstance(m1_value, np.ndarray):
                            # A large tolerance appears to be required for
                            # gribapi 1.12, but not for 1.14.
                            self.assertArrayAlmostEqual(
                                m1_value, m2_value, decimal=2
                            )
                        else:
                            self.assertEqual(
                                m1_value,
                                m2_value,
                                msg=msg.format(key, m1_value, m2_value),
                            )
                    else:
                        # We have a key which we expect to be different
                        # for each message.
                        self.assertEqual(
                            m1_value,
                            diffs[key][0],
                            msg=msg.format(key, m1_value, diffs[key][0]),
                        )

                        self.assertEqual(
                            m2_value,
                            diffs[key][1],
                            msg=msg.format(key, m2_value, diffs[key][1]),
                        )



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris_grib._save_rules.product_definition_template_10`

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock

from cf_units import Unit
import gribapi

from iris.coords import DimCoord
import iris.tests.stock as stock

from iris_grib._save_rules import product_definition_template_10


class TestPercentileValueIdentifier(tests.IrisGribTest):
    def setUp(self):
        self.cube = stock.lat_lon_cube()
        # Rename cube to avoid warning about unknown discipline/parameter.
        self.cube.rename('y_wind')
        time_coord = DimCoord(
            20, 'time', bounds=[0, 40],
            units=Unit('days since epoch', calendar='julian'))
        self.cube.add_aux_coord(time_coord)

    @mock.patch.object(gribapi, 'grib_set')
    def test_percentile_value(self, mock_set):
        cube = self.cube
        percentile_coord = DimCoord(95, long_name='percentile_over_time')
        cube.add_aux_coord(percentile_coord)

        product_definition_template_10(cube, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "productDefinitionTemplateNumber", 10)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "percentileValue", 95)

    @mock.patch.object(gribapi, 'grib_set')
    def test_multiple_percentile_value(self, mock_set):
        cube = self.cube
        percentile_coord = DimCoord([5, 10, 15],
                                    long_name='percentile_over_time')
        cube.add_aux_coord(percentile_coord, 0)
        err_msg = "A cube 'percentile_over_time' coordinate with one point "\
                  "is required"
        with self.assertRaisesRegex(ValueError, err_msg):
            product_definition_template_10(cube, mock.sentinel.grib)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris_grib._load_convert.data_representation_section.`

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from iris.exceptions import TranslationError

from iris_grib._load_convert import data_representation_section
from iris_grib.tests.unit import _make_test_message


class Test(tests.IrisGribTest):
    def test_supported_templates(self):
        template_nums = [0, 1, 2, 3, 4, 40, 41, 42, 50, 51, 61]
        for template_num in template_nums:
            message = _make_test_message(
                {5: {'dataRepresentationTemplateNumber': template_num}})
            data_representation_section(message.sections[5])

    def test_unsupported_template(self):
        message = _make_test_message(
            {5: {'dataRepresentationTemplateNumber': 5}})
        err_msg = r'Template \[5\] is not supported'
        with self.assertRaisesRegex(TranslationError, err_msg):
            data_representation_section(message.sections[5])


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests for `iris_grib._load_convert.satellite_common`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from unittest import mock
import warnings

from iris.coords import AuxCoord
import numpy as np

from iris_grib.tests.unit.load_convert import empty_metadata

from iris_grib._load_convert import satellite_common


class Test(tests.IrisGribTest):
    def _check(self, factors=1, values=111):
        # Prepare the arguments.
        series = mock.sentinel.satelliteSeries
        number = mock.sentinel.satelliteNumber
        instrument = mock.sentinel.instrumentType
        section = {'NB': 1,
                   'satelliteSeries': series,
                   'satelliteNumber': number,
                   'instrumentType': instrument,
                   'scaleFactorOfCentralWaveNumber': factors,
                   'scaledValueOfCentralWaveNumber': values}

        # Call the function.
        metadata = empty_metadata()
        satellite_common(section, metadata)

        # Check the result.
        expected = empty_metadata()
        coord = AuxCoord(series, long_name='satellite_series', units=1)
        expected['aux_coords_and_dims'].append((coord, None))
        coord = AuxCoord(number, long_name='satellite_number', units=1)
        expected['aux_coords_and_dims'].append((coord, None))
        coord = AuxCoord(instrument, long_name='instrument_type', units=1)
        expected['aux_coords_and_dims'].append((coord, None))
        standard_name = 'sensor_band_central_radiation_wavenumber'
        coord = AuxCoord(values / (10.0 ** factors),
                         standard_name=standard_name,
                         units='m-1')
        expected['aux_coords_and_dims'].append((coord, None))
        self.assertEqual(metadata, expected)

    def test_basic(self):
        self._check()

    def test_multiple_wavelengths(self):
        # Check with multiple values, and several different scaling factors.
        values = np.array([1, 11, 123, 1975])
        for i_factor in (-3, -1, 0, 1, 3):
            factors = np.ones(values.shape) * i_factor
            self._check(values=values, factors=factors)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function
:func:`iris_grib._load_convert.product_definition_template_0`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from unittest import mock

import iris.coords

import iris_grib
from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib.tests.unit.load_convert import LoadConvertTest
from iris_grib._load_convert import _MDI as MDI

from iris_grib._load_convert import product_definition_template_0


def section_4():
    return {'hoursAfterDataCutoff': MDI,
            'minutesAfterDataCutoff': MDI,
            'indicatorOfUnitOfTimeRange': 0,  # minutes
            'forecastTime': 360,
            'NV': 0,
            'typeOfFirstFixedSurface': 103,
            'scaleFactorOfFirstFixedSurface': 0,
            'scaledValueOfFirstFixedSurface': 9999,
            'typeOfSecondFixedSurface': 255}


class Test(LoadConvertTest):
    def test_given_frt(self):
        metadata = empty_metadata()
        rt_coord = iris.coords.DimCoord(24, 'forecast_reference_time',
                                        units='hours since epoch')
        product_definition_template_0(section_4(), metadata, rt_coord)
        expected = empty_metadata()
        aux = expected['aux_coords_and_dims']
        aux.append((iris.coords.DimCoord(6, 'forecast_period', units='hours'),
                    None))
        aux.append((
            iris.coords.DimCoord(30, 'time', units='hours since epoch'), None))
        aux.append((rt_coord, None))
        aux.append((iris.coords.DimCoord(9999, long_name='height', units='m'),
                    None))
        self.assertMetadataEqual(metadata, expected)

    def test_given_t(self):
        metadata = empty_metadata()
        rt_coord = iris.coords.DimCoord(24, 'time',
                                        units='hours since epoch')
        product_definition_template_0(section_4(), metadata, rt_coord)
        expected = empty_metadata()
        aux = expected['aux_coords_and_dims']
        aux.append((iris.coords.DimCoord(6, 'forecast_period', units='hours'),
                    None))
        aux.append((
            iris.coords.DimCoord(18, 'forecast_reference_time',
                                 units='hours since epoch'), None))
        aux.append((rt_coord, None))
        aux.append((iris.coords.DimCoord(9999, long_name='height', units='m'),
                    None))
        self.assertMetadataEqual(metadata, expected)

    def test_generating_process_warnings(self):
        metadata = empty_metadata()
        rt_coord = iris.coords.DimCoord(24, 'forecast_reference_time',
                                        units='hours since epoch')
        convert_options = iris_grib._load_convert.options
        emit_warnings = convert_options.warn_on_unsupported
        try:
            convert_options.warn_on_unsupported = True
            with mock.patch('warnings.warn') as warn:
                product_definition_template_0(section_4(), metadata, rt_coord)
            warn_msgs = [call[1][0] for call in warn.mock_calls]
            expected = ['Unable to translate type of generating process.',
                        'Unable to translate background generating process '
                        'identifier.',
                        'Unable to translate forecast generating process '
                        'identifier.']
            self.assertEqual(warn_msgs, expected)
        finally:
            convert_options.warn_on_unsupported = emit_warnings


if __name__ == '__main__':
    tests.main()
